from typing import List, Optional

import tsfel

from .extractor import BaseTSExtractor


class TSFEL(BaseTSExtractor):
    def __init__(
        self,
        domains: Optional[List[str]] = None,
        verbose: bool = False,
    ):
        # Generate default domain value
        """Initializes the TSFEL feature extractor with the specified domains
        and verbosity.

        Parameters
        ----------
        domains : Optional[List[str]]
            A list of domains to extract features from. If None, the default domains
            ["temporal", "statistical", "spectral"] will be used.
        verbose : bool
            If True, enables verbose output during feature extraction.
        """

        if domains is None:
            domains = ["temporal", "statistical", "spectral"]
        self.domains = domains
        self.verbose = verbose

        # update domain configurations
        self.cfg_file = {}
        for domain in self.domains:
            self.cfg_file.update(tsfel.get_features_by_domain(domain))

    def __call__(self, batch_windows, fs, dims):
        """Extracts features from a batch of samples.

        Parameters
        ----------
        batch_windows: List
            Batch of signals with len(dims) chans.
        fs: int
            Sampling frequency
        dims: List(str)
           list with the names of each signal dimension/channel ex: name of each ECG Lead

        Returns
        -------
        features: DataFrame
            DataFrame with the features from each batch.
        """
        features = tsfel.time_series_features_extractor(
            self.cfg_file,
            batch_windows,
            fs=fs,
            window_size=None,
            header_names=dims,
            verbose=int(self.verbose),
        )

        return features



import os
import sys
from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple, Union

import numpy as np
from loguru import logger
from torch.utils.data import DataLoader

from pymdma.common.definitions import InputLayer
from pymdma.constants import ReferenceType, ValidationDomain

from .data.load import TabularDataset

# Get the absolute path of the parent directory
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

# Add the parent directory to the Python path if it's not already included
if parent_dir not in sys.path:
    sys.path.insert(0, parent_dir)

SUPPORTED_FILES = {".json", ".xml", ".csv", ".xlsx"}  # TODO might want to add others or change


def _custom_collate_fn(batch):
    """Custom collate function to split batches into several elements.

    Parameters
    ----------
    batch : list of tuples
        A list where each element is a tuple containing encoded data, scaled data,
        column names, column map, quasi-identifier names, and sensitive names.

    Returns
    -------
    np.ndarray
        An array containing the scaled data extracted from the batch.
    """

    # custom collate function to handle varying sizes
    data_s = np.array(batch).astype(float)

    return data_s


def _get_data_files_path(data_src: Union[List[str], Path]) -> List[Path]:
    """Get a list of data files from the given source directory or list of file
    paths.

    Parameters
    ----------
    data_src : Union[List[str], Path]
        A list of file paths or a directory path.

    Returns
    -------
    List[Path]
        A list of data file paths.

    Raises
    ------
    ValueError
        If an invalid item is encountered or if an unsupported file extension is found.
    """

    if isinstance(data_src, list):
        data_files_path = data_src
    elif not data_src.is_dir():
        data_files_path = []
    else:
        data_files_path = []
        for item in data_src.iterdir():
            assert item.is_file() or item.is_dir(), f"Invalid item encountered: {item}"
            if item.is_file():
                assert item.suffix in SUPPORTED_FILES, f"Unsupported file extension: {item.suffix} (file: {item})"
                data_files_path.append(item)
            elif item.is_dir():
                # Recursively search for data files in subdirectories
                for data_file in item.iterdir():
                    if data_file.is_file() and data_file.suffix in SUPPORTED_FILES:
                        data_files_path.append(data_file)
                    else:
                        logger.warning(f"Unsupported file extension: {data_file.suffix} (file: {data_file})")
                        continue

    # tabular data usually only has one path
    data_path = data_files_path[0] if data_files_path else data_files_path

    return data_path


class TabularInputLayer(InputLayer):
    """Abstraction layer for handling different types of input data with the
    requirements from the data auditing module.

    Parameters
    ----------
    validation_domain : ValidationDomain
        The type of validation (input or synthetic).
    reference_type : ReferenceType
        The type of reference data (dataset or single tabular set).
    target_data : Path
        The path to the target dataset.
    reference_data : Optional[Path], optional
        The path to the reference dataset (used in synthetic evaluation).
    device : str, optional
        The device to be used in the feature extraction module (default is "cpu").
    scaler : str, optional
        The type of data normalization applied (default is None).
    embed : str, optional
        The type of embeddings for encoding features (default is None).
    **kwargs : optional
        Additional arguments passed to the InputLayer class.
    """

    def __init__(
        self,
        validation_domain: ValidationDomain,
        reference_type: ReferenceType,
        target_data: Path,
        reference_data: Optional[Path] = None,
        device: str = "cpu",
        scaler: str = None,
        embed: str = None,
        **kwargs,
    ) -> None:
        """Initializes the input layer with the given parameters.

        Parameters
        ----------
        validation_domain : ValidationDomain
            The validation type (input or synthetic).
        reference_type : ReferenceType
            The reference type (dataset or single tabular set).
        target_data : Path
            The path to the target dataset.
        reference_data : Optional[Path], optional
            The path to the reference dataset (default is None).
        device : str
            The device to be used in the feature extraction module (default is "cpu").
        scaler : str, optional
            The type of data normalization applied (default is None).
        embed : str, optional
            The type of embeddings for encoding features (default is None).
        **kwargs : optional
            Additional arguments passed to the InputLayer class.
        """

        super().__init__()
        self.val_type = validation_domain
        self.reference_type = reference_type
        self.device = device

        # scaler type
        self.scaler = scaler

        # embed type
        self.embed = embed

        # prepare reference dataloader (original/reference records)
        # will also be used for input validation
        if validation_domain == ValidationDomain.SYNTH:
            reference_file = _get_data_files_path(reference_data)
            reference_dataset = TabularDataset(
                file_path=reference_file,
                scaler=self.scaler,
                embed=self.embed,
                **kwargs,
            )
            self.reference_loader = DataLoader(
                dataset=reference_dataset,
                batch_size=len(reference_dataset),
                collate_fn=_custom_collate_fn,
            )

            # assign fitted scaler and embedders
            self.embed = reference_dataset.embed
            self.scaler = reference_dataset.scaler

        # prepare target dataloader (synthetic/similar records)
        # Will be used for synthetic evaluation or full reference input validation
        target_file = _get_data_files_path(target_data)
        target_dataset = TabularDataset(
            file_path=target_file,
            scaler=self.scaler,  # use scaler fitted by reference
            embed=self.embed,  # use embedder fitted by reference
            **kwargs,
        )
        self.target_loader = DataLoader(
            dataset=target_dataset,
            batch_size=len(target_dataset),
            collate_fn=_custom_collate_fn,
        )

        # TODO review for data modality
        # Client input validation
        if validation_domain == ValidationDomain.INPUT and reference_type != ReferenceType.NONE:
            assert len(reference_dataset) == len(
                target_dataset,
            ), "Reference and target datasets must have the same size for input validation."

    def __len__(self):
        """Returns the number of samples in the target dataset.

        Returns
        -------
        int
            The number of samples in the target dataset.
        """
        return len(self.target_loader.dataset)

    def get_embeddings(
        self,
        extractor_name: Optional[str] = None,
        model_instances: Optional[any] = None,
        **kwargs,
    ):
        """Retrieves the embeddings from the reference and target datasets.

        Parameters
        ----------
        extractor_name : Optional[str], optional
            The name of the extractor to use (default is None).
        model_instances : Optional[any], optional
            Additional model instances to use (default is None).
        **kwargs : optional
            Additional arguments for the extraction process.

        Returns
        -------
        tuple
            A tuple containing the embeddings from the reference and target datasets.
        """
        return self.reference_loader.dataset.data_emb, self.target_loader.dataset.data_emb

    @property
    def data_properties(self) -> Dict[str, any]:
        """Returns the properties of the target dataset.

        Returns
        -------
        Dict[str, any]
            A dictionary containing the properties of the target dataset.
        """
        return self.target_loader.dataset.properties

    @property
    def batched_samples(self) -> Generator[Tuple[np.ndarray], None, None]:
        """Yields batches of samples from the reference and target loaders.

        Yields
        ------
        Tuple[np.ndarray]
            A tuple containing the samples from the reference and target loaders.
        """

        # yielding batch of samples for evaluation
        if self.reference_type == ReferenceType.NONE:
            for sample in self.target_loader:
                yield (sample,)
        else:
            yield from zip(self.reference_loader, self.target_loader)



from typing import Union

import numpy as np

from pymdma.common.definitions import Metric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType


class Uniqueness(Metric):
    """Computes the percentage of consecutive equal values in the input
    signals. For multidimensional input signals, it considers the average of
    the consecutive values across the signal dimensions (e.g., leads).

    **Objective**: Uniqueness

    Parameters
    ----------
    tolerance : float, optional, default=0.0001
        The tolerance level for considering consecutive values as equal.
        A smaller tolerance means stricter equality. Defaults to 0.0001.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    Examples
    --------
    >>> uniqueness = Uniqueness()
    >>> sigs = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> result: MetricResult = uniqueness.compute(sigs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 100

    def __init__(
        self,
        tolerance: float = 0.0001,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.tolerance = tolerance

    def compute(self, signals: Union[np.ndarray, list], **kwargs) -> MetricResult:
        """Computes the ratio of consecutive equal values in the input signals.
        For multidimensional input signals, it considers the average of the
        consecutive values across the signal dimensions (e.g., leads).

        Parameters
        ----------
        signals : list or ndarray of shape (batch_size, signal_len, nr_dims)
            A batch of signals.

        Returns
        -------
        MetricResult
            A list of shape (batch_size) containing the percentage of consecutive values in each signal.
        """
        signals_np = np.array(signals)

        differences = np.abs(signals_np[:, :-1, :] - signals_np[:, 1:, :])

        within_tolerance = differences <= self.tolerance

        consecutive_equal_values = list(np.mean(within_tolerance, axis=(1, 2)))

        return MetricResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": consecutive_equal_values},
        )


class SNR(Metric):
    """The signal-to-noise ratio of the input data computed as the average of
    the mean to standard deviation ratio across signal dimensions.

    Parameters
    ----------
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Smith, S. W., The Scientist and Engineer's Guide to Digital Signal Processing (1997).
    https://dl.acm.org/doi/10.5555/281875

    Examples
    --------
    >>> snr = SNR()
    >>> sigs = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> result: MetricResult = snr.compute(sigs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = np.inf

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def compute(self, signals: Union[np.ndarray, list], **kwargs) -> MetricResult:
        """The signal-to-noise ratio of the input data computed as the average
        of the mean to standard deviation ratio across signal dimensions.

        Parameters
        ----------
        signals : list or ndarray of shape (batch_size, signal_len, nr_dims)
            A batch of signals.

        Returns
        -------
        MetricResult
            List of shape (batch_size) with the signal-to-noise ratio of each signal.
        """
        snrs = []
        for signal in signals:
            signal_mean = np.mean(signal, axis=0)
            signal_std = np.std(signal, axis=0)

            # Compute SNR using standard deviation method
            snr = np.mean(signal_mean / signal_std)
            snrs.append(snr)

        return MetricResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": snrs},
        )


__all__ = ["Uniqueness", "SNR"]



from pathlib import Path
from typing import List, Optional, Union, Tuple, Callable

import numpy as np
import pandas as pd
from torch.utils.data import Dataset

from ..data.harmonize import MetaEncoder
from ..embeddings.embed import GlobalEmbedder
from ..embeddings.scale import GlobalNorm
from ..embeddings.impute import GlobalImputer


def _read_xml(path: Path, **kwargs):
    """Read an XML file and converts it into a pandas DataFrame.

    Parameters
    ----------
    path : Path
        The path to the XML file.
    **kwargs : optional
        Additional arguments passed to pandas' read_xml function.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame containing the data read from the XML file, or None if an error occurs.
    """

    try:
        table = pd.read_xml(path, **kwargs)
    except Exception as e:
        table = None
    return table


def _read_json(path, **kwargs):
    """Read a JSON file converts it into a pandas DataFrame.

    Parameters
    ----------
    path : str
        The path to the JSON file.
    **kwargs : optional
        Additional arguments passed to pandas' read_json function.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame containing the data read from the JSON file, or None if an error occurs.
    """

    try:
        table = pd.read_json(path, orient="table", **kwargs)
    except Exception as e:
        table = None
    return table


def _read_xlsx(path: Path, **kwargs):
    """Read an Excel file and converts it into a pandas DataFrame.

    Parameters
    ----------
    path : Path
        The path to the Excel file.
    **kwargs : optional
        Additional arguments passed to pandas' read_excel function.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame containing the data read from the Excel file, or None if an error occurs.
    """

    try:
        table = pd.read_excel(path, **kwargs)
    except Exception as e:
        table = None
    return table


def _read_csv(path: Path, **kwargs):
    """Read a CSV file and converts it into a pandas DataFrame.

    Parameters
    ----------
    path : Path
        The path to the CSV file.
    **kwargs : optional
        Additional arguments passed to pandas' read_csv function.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame containing the data read from the CSV file, or None if an error occurs.
    """

    try:
        table = pd.read_csv(path, **kwargs)
    except Exception as e:
        table = None
    return table


class TabularDataset(Dataset):
    """
    A class for handling tabular datasets, including loading, encoding, and scaling.

    Parameters
    ----------
    file_path : Path or Callable
        The path to the dataset file or a callable for loading data.
    data : pd.DataFrame, optional
        A DataFrame containing the dataset. If provided, it bypasses loading from file.
    tag_name : str, optional
        A string to tag or label the dataset (e.g., 'real', 'synthetic').
    qi_names : list of str, optional
        List of quasi-identifier column names.
    sens_names : list of str, optional
        List of sensitive attribute column names.
    scaler : str or object, optional
        The scaling method or a scaler object to be used for data normalization.
    embed : str or object, optional
        The embedding method or an embedder object to be used for dimensionality reduction.
    meta : object, optional
        An instance of a meta encoder for variable type inference.
    scaler_kwargs : dict, optional
        Additional arguments for the scaler.
    embed_kwargs : dict, optional
        Additional arguments for the embedder.
    meta_kwargs : dict, optional
        Additional arguments for the meta encoder.
    with_onehot : bool, optional
        Whether to apply one-hot encoding to categorical variables.
    init_proc : bool, optional
        Whether to perform initial data processing.
    **kwargs : optional
        Additional arguments passed to the Dataset class.
    """
    
    LOAD_MAP = {
        "xml": _read_xml,
        "json": _read_json,
        "xlsx": _read_xlsx,
        "csv": _read_csv,
    }

    EMBED_AVAIL = [
        "umap",
        "pca",
        "lle",
        "isomap",
    ]

    SCALE_AVAIL = [
        "identity",
        "standard",
        "maxabs",
        "minmax",
        "robust",
    ]

    def __init__(
        self,
        file_path: Optional[Union[Path, str]],
        data: Optional[pd.DataFrame] = None,
        tag_name: Optional[str] = 'real',
        qi_names: Optional[str] = None,
        sens_names: Optional[str] = None,
        scaler: Optional[str] = None,
        embed: Optional[str] = None,
        imputer: Optional[str] = None,
        meta: Optional[object] = None,
        scaler_kwargs: Optional[dict] = {},
        embed_kwargs: Optional[dict] = {},
        imputer_kwargs: Optional[dict] = {},
        meta_kwargs: Optional[dict] = {},
        with_onehot: Optional[bool] = False,
        init_proc: Optional[bool] = True,
        **kwargs,
    ):
        """
        Initializes the TabularDataset with data loading, encoding, and scaling options.

        Parameters
        ----------
        file_path : Path
            The path to the dataset file.
        data : pd.DataFrame, optional
            A DataFrame containing the dataset. If provided, it bypasses loading from file.
        qi_names : list of str, optional
            List of quasi-identifier column names.
        sens_names : list of str, optional
            List of sensitive attribute column names.
        scaler : str or object, optional
            The scaling method or a scaler object to be used for data normalization.
        embed : str or object, optional
            The embedding method or an embedder object to be used for dimensionality reduction.
        meta : object, optional
            An instance of a meta encoder for variable type inference.
        scaler_kwargs : dict, optional
            Additional arguments for the scaler.
        embed_kwargs : dict, optional
            Additional arguments for the embedder.
        meta_kwargs : dict, optional
            Additional arguments for the meta encoder.
        with_onehot : bool, optional
            Whether to apply one-hot encoding to categorical variables.
        **kwargs : optional
            Additional arguments passed to the Dataset class.
        """
        
        # file path
        self.fpath = file_path

        # one hot encoding option
        self.oh_flag = with_onehot
        
        # dataset tag name
        self.tag = tag_name

        # metadata encoder
        self.col_map = None
        if meta is None:
            self.meta = MetaEncoder(**meta_kwargs)
            self.fit_meta = True
        else:
            self.meta = meta
            self.fit_meta = False

        # scaler
        if isinstance(scaler, (type(None), str)):
            self.scaler = GlobalNorm(n_type=scaler, **scaler_kwargs)
            self.fit_sc = True
        elif isinstance(scaler, object):
            self.scaler = scaler
            self.fit_sc = False
        else:
            self.scaler, self.fit_sc = None, None

        # embedder
        if isinstance(embed, (type(None), str)):
            self.embed = GlobalEmbedder(n_type=embed, **embed_kwargs)
            self.fit_emb = True
        elif isinstance(embed, object):
            self.embed = embed
            self.fit_emb = False
        else:
            self.embed, self.fit_emb = None, None

        # imputer
        if isinstance(imputer, (type(None), str)):
            self.imputer = GlobalImputer(n_type=imputer, **imputer_kwargs)
            self.fit_imp = True
        elif isinstance(imputer, object):
            self.imputer = imputer
            self.fit_imp = False
        else:
            self.imputer, self.fit_imp = None, None

        # data
        self.data = data
        self.data_enc = None
        self.data_s = None
        self.data_emb = None
        self.cols = None
        self.col_inds = None

        # special names
        self.qi_names, self.qi_cols = qi_names, None
        self.sens_names, self.sens_cols = sens_names, None

        # include data processing in init
        if init_proc:
            self.setup(self.data)

    def setup(self, data: Optional[pd.DataFrame] = None):
        # data transformations
        self.data, self.data_enc, self.data_s, self.cols = self.transform(
            path=self.fpath,
            data=data,
            scale_fit=self.fit_sc,
            meta_fit=self.fit_meta,
        )

        # data embeddings
        self.data_emb = self.embed_encode(
            data=self.data_s,
            with_fit=self.fit_emb,
            column_names=self.cols,
        )

        # special attributes
        self.qi_cols = self.get_special_columns(self.qi_names)  # quasi-identifiers
        self.sens_cols = self.get_special_columns(self.sens_names)  # sensitive attributes

    @property
    def properties(self):
        """
        Returns the properties of the dataset including column names,
        quasi-identifiers, sensitive attributes, and column mappings.

        Returns
        -------
        dict
            A dictionary containing the dataset properties.
        """
        
        return {
            "column_names": self.cols,
            "qi_names": self.qi_cols,
            "sens_names": self.sens_cols,
            "col_map": self.col_map,
        }

    def __len__(self):
        """
        Returns the number of rows in the dataset.

        Returns
        -------
        int
            The number of rows in the dataset.
        """
        
        return len(self.data)

    def __getcols__(self):
        """
        Returns the column names of the dataset provided.

        Returns
        -------
        list
            A list of column names in the dataset.
        """
        
        return self.cols

    def __getitem__(self, idx):
        """
        Returns the target data to use at a specific index, whether it is encoded, scaled, or embedded.

        Parameters
        ----------
        idx : int
            The index of the data to retrieve.

        Returns
        -------
        array
            An array containing target data at the specified index.
        """
        return self.data_s[idx]

    def __getdata__(self):
        """
        Returns the target data to use whether it is encoded, scaled, or embedded.

        Parameters
        ----------

        Returns
        -------
        array
            An array containing the whole target dataset.
        """
        return self.data_s
    
    def get_data(self):
        return self.__getdata__()
    
    def get_cols_inds(self):
        return self.col_inds

    def get_special_columns(self, names: List[str] = None):
        """
        Retrieves specified special columns based on provided names.

        Parameters
        ----------
        names : list of str, optional
            A list of names of the special columns to retrieve.

        Returns
        -------
        list
            A list of special columns found in the dataset.
        """
        
        # get special columns
        if names is not None:
            filt_names = []
            # check if names provided are acceptable real column names
            for name in names:
                find_ = np.char.find(self.cols, name) + 1
                if sum(find_ > 0) > 0:
                    filt_names.append(name)
            if not filt_names:
                filt_names = self.cols
        else:
            filt_names = self.cols

        return filt_names

    def meta_encode(
        self,
        data: np.ndarray,
        column_names: list = None,
        with_fit: bool = True,
        **kwargs,
    ):
        """
        Encodes the dataset using the meta encoder.

        Parameters
        ----------
        data : np.ndarray
            The input data to be encoded.
        column_names : list, optional
            The names of the columns to encode.
        with_fit : bool, optional
            Whether to fit the meta encoder before encoding.

        Returns
        -------
        np.ndarray
            The encoded dataset.
        """
        
        # infer variable dtypes from tabular data provided
        if with_fit:
            _ = self.meta.metadata(
                data=data,
                cols=column_names,
                **kwargs,
            )

        # encode data
        data_enc = self.meta.encode(
            data=data,
            column_names=column_names,
            **kwargs,
        )

        return data_enc

    def meta_decode(
        self,
        data: np.ndarray,
        columns_names: list = None,
        **kwargs,
    ):
        """
        Decodes the dataset using the meta encoder.

        Parameters
        ----------
        data : np.ndarray
            The encoded data to be decoded.
        columns_names : list, optional
            The names of the columns to decode.

        Returns
        -------
        np.ndarray
            The decoded dataset.
        """
        
        return self.meta.decode(data=data, column_names=columns_names, **kwargs)

    def scale_encode(
        self,
        data: np.ndarray,
        cols: List[int] = None,
        with_fit: bool = True,
        **kwargs,
    ):
        """
        Normalizes the dataset using the specified scaler.

        Parameters
        ----------
        data : np.ndarray
            The input data to be normalized.
        with_fit : bool, optional
            Whether to fit the scaler to the data before transforming.

        Returns
        -------
        np.ndarray
            The normalized dataset.
        """
        # copy of the data
        data_s = np.copy(data)

        # normalize data provided
        if with_fit:
            if cols is not None:
                data_s[:, cols] = self.scaler.fit_transform(data[:, cols], **kwargs)
            else:
                data_s = self.scaler.fit_transform(data, **kwargs)
        else:
            if cols is not None:
                data_s[:, cols] = self.scaler.transform(data[:, cols], **kwargs)
            else:
                data_s = self.scaler.transform(data, **kwargs)

        return data_s.astype(float)

    def scale_decode(
        self,
        data: np.ndarray,
        cols: List[int] = None,
        **kwargs,
    ):
        """
        Inverses the normalization on the dataset.

        Parameters
        ----------
        data : np.ndarray
            The normalized data to be inversely transformed.

        Returns
        -------
        np.ndarray
            The original dataset after inverse normalization.
        """
        # copy of the data
        data_cp = np.copy(data)

        # inverse transform
        if cols is not None:
            data_cp[:, cols] = self.scaler.inverse_transform(data[:, cols])
        else:
            data_cp = self.scaler.inverse_transform(data)

        return data_cp

    def embed_encode(
        self,
        data: np.ndarray,
        with_fit: bool = True,
        **kwargs,
    ):
        """
        Transforms the dataset using the specified embedding method.

        Parameters
        ----------
        data : np.ndarray
            The input data to be embedded.
        with_fit : bool, optional
            Whether to fit the embedder to the data before transforming.

        Returns
        -------
        np.ndarray
            The embedded dataset.
        """
        
        # if fit is needed before transforming
        if with_fit:
            data_emb = self.embed.fit_transform(data, **kwargs)
        
        # otherwise
        else:
            data_emb = self.embed.transform(data, **kwargs)

        return data_emb

    def embed_decode(
        self,
        data: np.ndarray,
        **kwargs,
    ):
        """
        Reverses the transformation on the dataset.

        Parameters
        ----------
        data : np.ndarray
            The embedded data to be inversely transformed.

        Returns
        -------
        np.ndarray
            The original dataset after inverse embedding.
        """
        
        return self.embed.inverse_transform(data)

    def read_data(
        self,
        path: Optional[Union[Path, str]] = None,
        data: pd.DataFrame = None,
        **kwargs,
    ):
        """
        Reads the dataset from a file or returns the provided DataFrame.

        Parameters
        ----------
        path : Path, optional
            The path to the dataset file.
        data : pd.DataFrame, optional
            A DataFrame to return if provided.

        Returns
        -------
        pd.DataFrame or np.ndarray
            The loaded dataset as a DataFrame or numpy array.
        """
        
        # data reading
        if isinstance(data, (pd.DataFrame, np.ndarray)):
            data_r, tgt_r = data, None

        elif isinstance(path, (str, Path)):
            fmt = Path(path).suffix[1:].lower()
            loader = self.LOAD_MAP.get(fmt)

            # get data
            data_r = loader(path, **kwargs) if loader is not None else None
            tgt_r = None

        elif isinstance(path, Callable):
            data_r, tgt_r = path()

        else:
            data_r, tgt_r = None, None

        return data_r, tgt_r
    
    def impute_missing(self, data: np.ndarray, cols: List[int] = None, miss_inds: List[Tuple[int, int]] = None, meta_fit: bool = True, **kwargs):
        # meta encoding
        _ = self.meta_encode(
            data=data,
            column_names=cols,
            with_fit=meta_fit,
            with_onehot=False,
        )

        # for imputation
        data_ = self.meta_encode(
            data=data,
            column_names=cols,
            with_fit=False,
            with_onehot=False,
        )

        # replace back missing values
        if miss_inds:
            data_[tuple(zip(*miss_inds))] = np.nan
        
        # impute missing values
        imp = GlobalImputer(n_type='knn', norm_type='stardard', **kwargs)
        data_imp = imp.fit_transform(data_)

        # decode back
        data_dec = self.meta_decode(
            data=data_imp, 
            columns_names=cols, 
            with_onehot=False
        )

        # reset metadata params
        if meta_fit:
            self.meta.reset_params()

        return data_dec

    def transform(
        self,
        path: Path = None,
        data: pd.DataFrame = None,
        scale_fit: bool = True,
        meta_fit: bool = True,
        **kwargs,
    ):
        """
        Loads, encodes, and scales the dataset.

        Parameters
        ----------
        path : Path, optional
            The path to the dataset file.
        data : pd.DataFrame, optional
            A DataFrame to be processed if provided.
        scale_fit : bool, optional
            Whether to fit the scaler to the data.
        meta_fit : bool, optional
            Whether to fit the meta encoder to the data.

        Returns
        -------
        tuple
            A tuple containing the original data, encoded data, scaled data, and column names.
        """
        
        # data loader
        data_r, _ = self.read_data(path, data, **kwargs)

        # get missing indices
        miss_inds = list(zip(*np.where(data_r.isna().to_numpy())))

        # replace missing values
        data_r = data_r.fillna(data_r.mode().iloc[0])
        
        # check if data was correctly loaded
        if data_r is not None:
            # get columns/attributes
            cols = list(data_r.columns)

            # numpy array
            data_np = data_r.to_numpy()

            # impute missing values
            data_imp = self.impute_missing(
                data=data_np,
                cols=cols,
                miss_inds=miss_inds,
                meta_fit=meta_fit
            )

            # meta encoding
            data_enc = self.meta_encode(
                data=data_imp,
                column_names=cols,
                with_fit=meta_fit,
                with_onehot=self.oh_flag,
            )

            # column map
            self.col_map = self.meta.vtype_map

            # column indices
            or_inds = self.meta.or_inds
            self.col_inds = self.meta.col_inds  # assign col indices

            # norm scaling
            data_s = self.scale_encode(
                data=data_enc, 
                cols=or_inds, 
                with_fit=scale_fit
            )

        else:
            # default values
            data_r, data_enc, data_s, cols = [], [], [], []

        return data_r, data_enc, data_s, cols

    def inverse_transform(self, data: np.ndarray, column_names: list = None):
        """
        Applies the inverse transformations on the dataset.

        Parameters
        ----------
        data : np.ndarray
            The data to apply inverse transformations to.
        column_names : list, optional
            Full list of column names.

        Returns
        -------
        np.ndarray
            The dataset after applying inverse transformations.
        """
        
        # inverse scaling
        col_inds = self.meta.or_inds
        data = self.scale_decode(data, cols=col_inds)

        # meta decoding
        data = self.meta_decode(
            data=data,
            columns_names=column_names,
            with_onehot=self.oh_flag,
        )

        return data
    
if __name__ == '__main__':
    import pandas as pd
    
    # data
    data = pd.DataFrame({'a': [1,2,3,4,5], 'b': [1,2,3,4,5], 'c': [0.2,0.3,0.5,0.6,5.4], 'd': ['a','b','c','d','e'], 'e': [0,1,0,1,0]})

    # loader
    loader = TabularDataset(data=data, init_proc=False)

    # setup
    data = loader.setup(data=data)




from typing import List

import numpy as np
import pandas as pd
import scipy.stats as stat
from loguru import logger
from pycanon import anonymity
from sklearn.cluster import DBSCAN
from sklearn.metrics import silhouette_score
from sklearn.neighbors import NearestNeighbors
from statsmodels.stats.outliers_influence import variance_inflation_factor

# # DESCRIPTIVE STATISTICS
# def n_sample(data: np.ndarray):
#     """Computes the number of samples in the dataset.

#     Args:
#         data (np.ndarray): Input dataset.

#     Returns:
#         int: Number of samples.
#     """
#     return len(data)


# def mean_(data: np.ndarray, axis: int = 0, **kwargs):
#     """Computes the mean along the specified axis.

#     Args:
#         data (np.ndarray): Input data.
#         axis (int, optional): Axis along which the mean is computed. Defaults to 0.

#     Returns:
#         np.ndarray: Mean values.
#     """
#     return np.mean(data, axis=axis)


# def stdev_(data: np.ndarray, axis: int = 0, **kwargs):
#     """Computes the standard deviation along the specified axis.

#     Args:
#         data (np.ndarray): Input data.
#         axis (int, optional): Axis along which the standard deviation is computed. Defaults to 0.

#     Returns:
#         np.ndarray: Standard deviation values.
#     """
#     return np.std(data, axis=axis)


def percentile_(data: np.ndarray, percentile: int = 75, **kwargs):
    """Computes the specified percentile of the data.

    Args:
        data (np.ndarray): Input data.
        percentile (int, optional): Percentile to compute. Defaults to 75.

    Returns:
        np.ndarray: Percentile values.
    """
    return np.percentile(data, q=percentile)


# def max_(data: np.ndarray, axis: int = 0, **kwargs):
#     """Computes the maximum value along the specified axis.

#     Args:
#         data (np.ndarray): Input data.
#         axis (int, optional): Axis along which the maximum value is computed. Defaults to 0.

#     Returns:
#         np.ndarray: Maximum values.
#     """
#     return np.max(data, axis=axis)


# def min_(data: np.ndarray, axis: int = 0, **kwargs):
#     """Computes the minimum value along the specified axis.

#     Args:
#         data (np.ndarray): Input data.
#         axis (int, optional): Axis along which the minimum value is computed. Defaults to 0.

#     Returns:
#         np.ndarray: Minimum values.
#     """
#     return np.min(data, axis=axis)


# def skewness_(data: np.ndarray, axis: int = 0, **kwargs):
#     """Computes the skewness along the specified axis.

#     Args:
#         data (np.ndarray): Input data.
#         axis (int, optional): Axis along which the skewness is computed. Defaults to 0.

#     Returns:
#         float: Skewness values.
#     """
#     return stat.skewness(data, axis=axis)


# def kurtosis_(data: np.ndarray, axis: int = 0, **kwargs):
#     """Computes the kurtosis along the specified axis.

#     Args:
#         data (np.ndarray): Input data.
#         axis (int, optional): Axis along which the kurtosis is computed. Defaults to 0.

#     Returns:
#         float: Kurtosis values.
#     """
#     return stat.kurtosis(data, axis=axis)


# def cv_(data: np.ndarray, axis: int = 0, **kwargs):
#     """Computes the coefficient of variation along the specified axis.

#     Args:
#         data (np.ndarray): Input data.
#         axis (int, optional): Axis along which the coefficient of variation is computed. Defaults to 0.

#     Returns:
#         np.ndarray: Coefficient of variation values.
#     """
#     return np.divide(stdev_(data, axis), mean_(data, axis))


def entropy_(data: np.ndarray, axis: int = 0, **kwargs):
    """Computes the entropy along the specified axis.

    Args:
        data (np.ndarray): Input data.
        axis (int, optional): Axis along which the entropy is computed. Defaults to 0.

    Returns:
        float: Entropy values.
    """
    # number of unique values
    unq = len(data)

    # max possible entropy
    max_val = np.log2(unq)

    # compute normalised entropy
    if unq == 1:
        entpy = 1.0
    else:
        entpy = stat.entropy(data, axis=axis, base=2) / max_val

    return entpy


def chi2_uniformity(data, **kwargs):
    if len(data) > 1:
        # expected distribution
        expected = np.full_like(data, fill_value=np.mean(data))

        # perform test
        chi2, p_value = stat.chisquare(data, f_exp=expected)

        # norm
        chi2_norm = chi2 / (len(data) - 1)
    else:
        # norm
        chi2_norm = 0.0

    return chi2_norm


def ks_uniformity(data, **kwargs):
    # observed data
    obs_data = data.copy()

    # expected data
    exp_data = np.random.uniform(
        low=min(obs_data),
        high=max(obs_data),
        size=len(obs_data),
    )

    # perform test
    ks_stat, p_value = stat.kstest(obs_data, exp_data)

    return ks_stat


def stat_test_uniformity(data: np.ndarray, var: np.ndarray, tag: str, **kwargs):
    # score
    if tag == "discrete":
        score = 1 - chi2_uniformity(var, **kwargs)
    elif tag == "continuous":
        score = 1 - ks_uniformity(data, **kwargs)
    else:
        score = -1

    return score


def iqr_(data: np.ndarray, **kwargs):
    """Computes the interquartile range (IQR) of the data.

    Args:
        data (np.ndarray): Input data.

    Returns:
        np.ndarray: Interquartile range (IQR) values.
    """
    return percentile_(data, 75) - percentile_(data, 25)


# STATISTICAL OUTLIERS
def z_score_outliers(data: np.ndarray, threshold: int = 3, **kwargs):
    """Counts the number of outliers found through the z-score metric.

    Args:
        data (np.ndarray): Input data.
        threshold (int, optional): Z-score threshold for identifying outliers. Defaults to 3.

    Returns:
        int: Number of outliers.
    """
    z_scores = np.abs((data - data.mean()) / data.std())
    return np.sum(z_scores > threshold)


def iqr_outliers(data: np.ndarray, factor: float = 1.5, **kwargs):
    """Counts the number of outliers found through the interquartile range
    (IQR) method.

    Args:
        data (np.ndarray): Input data.
        factor (float, optional): IQR factor for identifying outliers. Defaults to 1.5.

    Returns:
        int: Number of outliers.
    """
    # Quartiles
    Q1 = percentile_(data, 25)
    Q3 = percentile_(data, 75)

    # Interquartile range
    IQR = Q3 - Q1

    # Outlier Boundaries
    lower_bound = Q1 - factor * IQR
    upper_bound = Q3 + factor * IQR

    return np.sum((data < lower_bound) | (data > upper_bound))


# def descriptive_stat(data: np.ndarray, cols: list = None):
#     # columns
#     cols_ = [f"attr{x}" for x in range(data.shape[1])] if cols is None else cols

#     # total count
#     n_total = len(data)

#     # valid count
#     n_valid = len(np.isfinite(data.astype(float)))

#     # mean
#     mean_ = np.round(np.mean(data, axis=0), 4)

#     # std
#     std_ = np.round(np.std(data, axis=0), 4)

#     # min
#     min_ = np.round(np.min(data, axis=0), 4)

#     # 25 perc
#     p_25 = np.round(np.percentile(data, 25, axis=0), 4)

#     # 50 perc
#     p_50 = np.round(np.percentile(data, 50, axis=0), 4)

#     # 75 perc
#     p_75 = np.round(np.percentile(data, 75, axis=0), 4)

#     # max
#     max_ = np.round(np.max(data, axis=0), 4)

#     # metrics stacked
#     metrics = {
#         "n_total": [n_total] * len(cols_),
#         "n_valid": [n_valid] * len(cols_),
#         "mean": mean_,
#         "std": std_,
#         "min": min_,
#         "25%": p_25,
#         "50%": p_50,
#         "75%": p_75,
#         "max": max_,
#     }

#     # stats dict
#     stat = {col: {f"{k}": v[idx] for k, v in metrics.items()} for idx, col in enumerate(cols_)}

#     return stat


def corr_matrix(data: np.ndarray, **kwargs):
    """Computes the correlation matrix for the input data.

    Args:
        data (np.ndarray): Input dataset.
        col_map (dict): Column mapper for encoding categorical attributes.
        corr_type (str, optional): Type of correlation to compute ('pearson', 'kendall', 'spearman'). Defaults to 'pearson'.

    Returns:
        np.ndarray: Correlation matrix.
    """

    # unique values
    unq = np.unique(data)

    # compute linear correlation matrix
    if len(unq) > 1:
        corr_m = np.corrcoef(data.astype(float), rowvar=False)
    else:
        corr_m = np.ones((data.shape[-1], data.shape[-1])).astype(float)

    return corr_m


def corr_strong(corr_matrix: np.ndarray, cols: list, c_thresh: float = 0.5, **kwargs):
    """Computes correlation statistics for each column based on a correlation
    threshold.

    Args:
        corr_matrix (pd.DataFrame): Correlation matrix.
        c_thresh (float, optional): Correlation threshold for identifying correlated columns. Defaults to 0.5.

    Returns:
        dict: Correlation statistics.
    """
    stats_d = {}

    # filter out weak relationships
    corr_ = np.array(corr_matrix.__abs__() > c_thresh)

    # numpy column array
    cols_ = np.array(cols)

    # check strong correlations per column
    for idx, col_ in enumerate(cols):
        # strong correlation indices
        corr_ind = np.where(corr_[:, idx])[0]
        corr_ind = corr_ind[corr_ind != idx]  # remove the target column

        # get strongly correlation columns and its values
        corr_cols = cols_[corr_ind]
        corr_vals = corr_matrix[idx, corr_ind]

        stats_d[col_] = {col: val for col, val in zip(corr_cols, corr_vals)}

    return stats_d


def uniformity_score_per_column(
    data_col: np.ndarray,
    is_continuous: bool,
    n_categories: int = None,
    **kwargs,
):
    """Detect target attribute imbalancing likelihood. This function computes
    the CV, Entropy and IQR metrics to assess an imbalancing score from the
    target attribute side. Rule of thumb: CV < [0.10, 0.15], IQR < 0.25*mean,
    entropy.

    Args:
        data (np.ndarray): _description_
        target_col (str): _description_

    Returns:
        _type_: _description_
    """

    # target attribute distribution
    if is_continuous:
        # create histogram for continuous data
        tag = "continuous"
        var, _ = np.histogram(data_col, bins="auto", density=True)
    else:
        # calculate counts for discrete data
        tag = "discrete"
        _, var = np.unique(data_col, return_counts=True)

    # normalise counts
    var = var / var.sum()

    # if specified, adjust for expected number of categories
    if isinstance(n_categories, int) and not is_continuous:
        n_diff = n_categories - len(var)
        var = np.append(var, [0] * n_diff)

    # calculate scores
    stat_score = stat_test_uniformity(data=data_col, var=var, tag=tag)
    entropy_score = round(entropy_(var), 4)

    # define flags based on heuristic thresholds
    stat_flag = stat_score > 0.80
    entropy_flag = entropy_score > 0.95

    # calculate imbalance level
    imb_level = round(100 * (1 - np.mean([stat_flag, entropy_flag])), 1)

    return stat_score, entropy_score, imb_level


# def dtype_inference(data: pd.DataFrame, **kwargs):
#     check_map = {"dtype": []}  # dtype validation map {col: check}

#     # Schema Validation
#     for col in data.columns:
#         check_map["dtype"].append(data[col].dtype.__str__())

#     return pd.DataFrame(check_map, index=data.columns).T


# def consistency_check(data: pd.DataFrame, categorical_map: dict, **kwargs):
#     # Consistency Check
#     for col, avail_categ in categorical_map.items():
#         unique_categories = data[col].unique()
#         predefined_categories = [unq in avail_categ for unq in unique_categories]
#         if not set(unique_categories).issubset(predefined_categories):
#             return False
#     return True


# def estimate_eps(embeddings: np.ndarray, k: int):
#     # Compute distances to k nearest neighbors
#     nbrs = NearestNeighbors(n_neighbors=k + 1).fit(embeddings)
#     distances, _ = nbrs.kneighbors(embeddings)

#     # Take the average of distances to k nearest neighbors
#     avg_distances = np.mean(distances[:, 1:], axis=1)

#     # Choose eps as a fraction of the average distance
#     eps = np.percentile(avg_distances, 90)  # Adjust the percentile as needed

#     return eps


# def proximity_score(embeddings: np.ndarray, n_neighbors: int = 4, **kwargs):
#     # estimate eps param for DBSCAN initialization
#     eps_ = estimate_eps(embeddings, k=n_neighbors)

#     # perform DBSCAN clustering
#     dbscan = DBSCAN(eps=eps_)
#     cluster_lbl = dbscan.fit_predict(embeddings)

#     # number of clusters found (-1 indicates noisy samples)
#     n_clusters = len(np.unique(cluster_lbl[cluster_lbl > -1]))

#     # compute silhouette score
#     if n_clusters > 1:
#         score = silhouette_score(
#             embeddings[cluster_lbl > -1],
#             cluster_lbl[cluster_lbl > -1],
#         )
#     else:
#         score = 1

#     return score


# GDPR privacy
def compute_k_anonymity(data: np.ndarray, column_names: List[str], qi_names: List[str] = None, **kwargs):
    # convert into dataframe
    aux_data = pd.DataFrame(data, columns=column_names)

    # compute k-anonimity
    if isinstance(qi_names, list):
        if len(qi_names) > 0:
            # get k-anom value
            k_anom = anonymity.k_anonymity(aux_data, qi_names)

            # convert into percentage (the higher the better privacy is preserved)
            k_anom = 100 * (1 - k_anom / len(aux_data))
        else:
            k_anom = -1
    else:
        k_anom = -1

    return float(k_anom)


# multicolinearity
def compute_vif(data: np.ndarray, column_names: List[str], **kwargs):
    vif_map, vif_val = {}, []

    for idx, col in enumerate(column_names):
        # column-wise vif
        aux_vif = variance_inflation_factor(data, idx)

        # assign vif
        vif_map[col] = aux_vif
        vif_val.append(aux_vif > 5)  # multicolinearity threshold

    # get percentage of columns above the threshold
    vif_perc = sum(vif_val) / len(vif_val)

    return vif_perc, vif_map



from typing import Sequence

import numpy as np
from loguru import logger

from pymdma.common.definitions import Metric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

# from https://github.com/scikit-image/scikit-image/blob/main/skimage/color/colorconv.py#L736
_yiq_from_rgb = np.array(
    [
        [0.299, 0.587, 0.114],
        [0.59590059, -0.27455667, -0.32134392],
        [0.21153661, -0.52273617, 0.31119955],
    ],
).T.astype(np.float32)


class PSNR(Metric):
    """Computes Peak Signal to Noise Ratio (PSNR) between two images.

    **Objective**: Signal-to-noise ratio

    Parameters
    ----------
    data_range : int, optional, default=255
        Maximum value of the image data (e.g., 255 for 8-bit images).
    convert_to_grayscale : bool, optional, default=False
        Whether to convert the images to grayscale before computing PSNR.
    allow_nan : bool, optional, default=False
        Whether to allow NaN values in the returned scores.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unsused).

    References
    ----------
    wikipedia.org: Peak signal-to-noise ratio
    https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    Examples
    --------
    >>> psnr = PSNR()
    >>> reference_imgs = np.random.rand(20, 256, 256, 3) # (N, H, W, C)
    >>> target_imgs = np.random.rand(20, 256, 256, 3) # (N, H, W, C)
    >>> result: MetricResult = psnr.compute(reference_imgs, target_imgs)
    """

    reference_type = ReferenceType.INSTANCE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = np.inf

    def __init__(
        self,
        data_range: int = 255,
        convert_to_grayscale: bool = False,
        allow_nan: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.data_range = data_range
        self.convert_to_grayscale = convert_to_grayscale
        self.allow_nan = allow_nan
        self._eps = 1e-8 if not allow_nan else 0

    def compute(
        self,
        reference_imgs: Sequence[np.ndarray],
        target_imgs: Sequence[np.ndarray],
        **kwargs,
    ) -> MetricResult:
        """Computes PSNR score between two image sets.

        Parameters
        ----------
        reference_imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            Images to use as reference.
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.
        target_imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            Corresponding images to compare with reference.
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.

        Returns
        -------
        result : MetricResult
            Instance-level PSNR scores.
            May contain warnings if PSNR is infinite.
        """
        assert len(reference_imgs) == len(target_imgs), "Reference and target images must have the same length"
        assert all(
            ref.shape == targ.shape for ref, targ in zip(reference_imgs, target_imgs)
        ), "Reference and target images must have the same shape"

        # convert images to YIQ and extract luminance
        if self.convert_to_grayscale:
            reference_imgs = [np.sum(ref @ _yiq_from_rgb, axis=-1) for ref in reference_imgs]
            target_imgs = [np.sum(targ @ _yiq_from_rgb, axis=-1) for targ in target_imgs]

        mses = np.array(
            [
                np.mean((ref.astype(np.float32) - targ.astype(np.float32)) ** 2)
                for ref, targ in zip(reference_imgs, target_imgs)
            ],
        )
        psnrs = 20 * np.log10(self.data_range / (np.sqrt(mses) + self._eps))

        warnings = set()
        if np.isinf(psnrs).any():
            warnings.add("Infinite PSNR value. Check if reference image is the same as the target image.")
            logger.warning("Infinite PSNR value. Check if reference image is the same as the target image.")

        return MetricResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": psnrs.tolist()},
            errors=list(warnings) if len(warnings) > 0 else None,
        )


__all__ = ["PSNR"]



from typing import Dict, Literal, Optional, Union

import numpy as np
import torch
from piq import FID as _FID
from piq import GS as _GS
from piq import MSID as _MSID
from torch import Tensor

from pymdma.common.definitions import FeatureMetric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

from ...functional.ratio import dispersion_ratio, distance_ratio
from ...utils.util import features_splitting, to_tensor


class FrechetDistance(FeatureMetric):
    """Frechet Distance (FD) metric wrapper from the PIQ implementation of FID.
    Allows the computation of the dispersion and distance ratios for the
    metric.

    **Objective**: Fidelity, Diversity

    Parameters
    ----------
    compute_ratios : bool, optional
        If set to True, the dispersion and distance ratios will be computed. Defaults to True.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unused).

    Notes
    -----
    This implementation is based on the PIQ library. The base extractor model is InceptionV3, but this implementation allows
    for the use of other embedding models (useful when the synthetic data is not compatible with Inception models).

    See Also
    --------
    general.functional.ratio.dispersion_ratio : Compute the dispersion ratio for the Frechet Distance metric.
    general.functional.ratio.distance_ratio : Compute the distance ratio for the Frechet Distance metric.

    References
    ----------
    Kastryulin et al., PyTorch Image Quality: Metrics for Image Quality Assessment (2022).
    https://arxiv.org/abs/2208.14818

    piq, PyTorch Image Quality: Metrics and Measure for Image Quality Assessment,
    https://github.com/photosynthesis-team/piq

    Examples
    --------
    >>> fid = FrechetDistance()
    >>> x_feats = np.random.rand(100, 100)
    >>> y_feats = np.random.rand(100, 100)
    >>> result: MetricResult = fid.compute(x_feats, y_feats)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = np.inf

    extractor_model_name: str = "inception_fid"

    def __init__(
        self,
        compute_ratios: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.compute_ratios = compute_ratios
        self._fid_inst = _FID()

    def _compute_ratios(self, real_features: np.ndarray, fake_features: np.ndarray, **kwargs) -> Dict[str, float]:
        """Compute the dispersion and distance ratios for the Frechet Distance
        metric.

        Parameters
        ----------
        real_features : np.ndarray
            Array of shape (n_samples, n_features) containing features of real samples.
        fake_features : np.ndarray
            Array of shape (n_samples, n_features) containing features of fake/generated samples.

        Returns
        -------
        ratios: dict
            Dictionary containing the dispersion and distance ratio values.
        """
        state = kwargs.get("context", {})

        if any(key not in state for key in {"x_split_1", "x_split_2", "y_split_1", "y_split_2"}):
            state["x_split_1"], state["x_split_2"] = features_splitting(real_features, seed=0)
            state["y_split_1"], state["y_split_2"] = features_splitting(fake_features, seed=0)

        return {
            "dispersion_ratio": dispersion_ratio(
                self._fid_inst.compute_metric,
                to_tensor(state["x_split_1"]),
                to_tensor(state["x_split_2"]),
                to_tensor(state["y_split_1"]),
                to_tensor(state["y_split_2"]),
            ),
            "distance_ratio": distance_ratio(
                self._fid_inst.compute_metric,
                to_tensor(state["x_split_1"]),
                to_tensor(state["x_split_2"]),
                to_tensor(state["y_split_1"]),
                to_tensor(state["y_split_2"]),
            ),
        }

    def compute(
        self,
        real_features: Union[Tensor, np.ndarray],
        fake_features: Union[Tensor, np.ndarray],
        **kwargs,
    ) -> MetricResult:
        """Compute the Frechet Distance metric.

        Parameters
        ----------
        real_features : Union[Tensor, np.ndarray]
            Array-like of shape (n_samples, n_features) containing features of real samples.
        fake_features : Union[Tensor, np.ndarray]
            Array-like of shape (n_samples, n_features) containing features of fake/generated samples.

        Returns
        -------
        result : MetricResult
            Dataset-level FD score and the dispersion and distance ratios.
        """
        real_features = to_tensor(real_features)
        fake_features = to_tensor(fake_features)

        fid = self._fid_inst.compute_metric(real_features, fake_features)

        ratios = None
        if self.compute_ratios:
            ratios = self._compute_ratios(real_features, fake_features, **kwargs)

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.NUMERIC,
                "subtype": "float",
                "value": fid.detach().item(),
                "stats": ratios,
            },
        )


class GeometryScore(FeatureMetric):
    """Geometry Score (GS) metric wrapper from the PIQ implementation of GS.

    **Objective**: Fidelity, Diversity

    Parameters
    ----------
    sample_size : int, optional
        Number of samples to use for the GS computation. Defaults to 128.
    num_iters : int, optional
        Number of iterations to use for the GS computation. Defaults to 1000.
    gamma : float, optional
        Gamma parameter for the GS computation. Defaults to None.
    i_max : int, optional
        Maximum number of iterations for the GS computation. Defaults to 10.
    num_workers : int, optional
        Number of workers to use for the GS computation. Defaults to 4.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unused).

    References
    ----------
    Kastryulin et al., PyTorch Image Quality: Metrics for Image Quality Assessment (2022).
    https://arxiv.org/abs/2208.14818

    piq, PyTorch Image Quality: Metrics and Measure for Image Quality Assessment,
    https://github.com/photosynthesis-team/piq

    Examples
    --------
    >>> gs = GeometryScore()
    >>> x_feats = np.random.rand(100, 100)
    >>> y_feats = np.random.rand(100, 100)
    >>> result: MetricResult = gs.compute(x_feats, y_feats)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = np.inf

    def __init__(
        self,
        sample_size: int = 128,
        num_iters: int = 1000,
        gamma: Optional[float] = None,
        i_max: int = 10,
        num_workers: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sample_size = sample_size
        self.num_iters = num_iters
        self.gamma = gamma
        self.i_max = i_max
        self.num_workers = num_workers

        self._gs = _GS(
            sample_size=self.sample_size,
            num_iters=self.num_iters,
            gamma=self.gamma,
            i_max=self.i_max,
            num_workers=self.num_workers,
        )

    def compute(
        self,
        real_features: Union[Tensor, np.ndarray],
        fake_features: Union[Tensor, np.ndarray],
        **kwargs,
    ) -> MetricResult:
        """Compute the Geometry Score metric.

        Parameters
        ----------
        real_features : Union[Tensor, np.ndarray]
            Array-like of shape (n_samples, n_features) containing features of real samples.
        fake_features : Union[Tensor, np.ndarray]
            Array-like of shape (n_samples, n_features) containing features of fake/generated samples.

        Returns
        -------
        result : MetricResult
            Dataset-level GS score.
        """
        real_features = to_tensor(real_features)
        fake_features = to_tensor(fake_features)

        score = self._gs.compute_metric(real_features, fake_features)

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": score.detach().item()},
        )


class MultiScaleIntrinsicDistance(FeatureMetric):
    """Multi-Scale Intrinsic Distance (MSID) metric wrapper from the PIQ
    implementation of MSID.

    **Objective**: Fidelity, Diversity

    Parameters
    ----------
    ts : Optional[Tensor], optional
        Tensor of shape (n_samples, n_features) containing the temperature values. Defaults to None.
    k_neighbours : int, optional
        Number of nearest neighbours to consider. Defaults to 5.
    m_steps : int, optional
        Number of steps for the MSID computation. Defaults to 10.
    niters : int, optional
        Number of iterations for the MSID computation. Defaults to 100.
    rademacher : bool, optional
        Whether to use Rademacher distribution for the MSID computation. Defaults to False.
        When not active will use standard normal for random vectors in Hutchinson.
    normalized_laplacian : bool, optional
        Whether to normalize the laplacian for the MSID computation. Defaults to True.
    normalize : Literal["empty", "complete", "er", "none"], optional
        Normalization strategy for the laplacian. Defaults to "empty".
    msid_mode : Literal["l2", "max"], optional
        Mode for the MSID computation. Defaults to "max".
    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unused).

    Notes
    -----
    The results of this metric are based on random approximations, so they are not deterministic.
    In some datasets the results can be unstable. This can be mitigated by increasing the
    number of iterations with the `niters` parameter.


    References
    ----------
    Kastryulin et al., PyTorch Image Quality: Metrics for Image Quality Assessment (2022).
    https://arxiv.org/abs/2208.14818

    piq, PyTorch Image Quality: Metrics and Measure for Image Quality Assessment,
    https://github.com/photosynthesis-team/piq

    Examples
    --------
    >>> msid = MultiScaleIntrinsicDistance()
    >>> x_feats = np.random.rand(100, 100)
    >>> y_feats = np.random.rand(100, 100)
    >>> result: MetricResult = msid.compute(x_feats, y_feats)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = np.inf

    extractor_model_name: str = "inception_fid"

    def __init__(
        self,
        ts: Optional[Union[Tensor, np.ndarray]] = None,
        k_neighbours: int = 5,
        m_steps: int = 10,
        niters: int = 100,
        rademacher: bool = False,
        normalized_laplacian: bool = True,
        normalize: Literal["empty", "complete", "er", "none"] = "empty",
        msid_mode: Literal["l2", "max"] = "max",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.ts = torch.from_numpy(ts) if isinstance(ts, np.ndarray) else ts
        self.k_neighbours = k_neighbours
        self.m_steps = m_steps
        self.niters = niters
        self.rademacher = rademacher
        self.normalized_laplacian = normalized_laplacian
        self.normalize = normalize
        self.msid_mode = msid_mode

        self._msid = _MSID(
            ts=self.ts,
            k=self.k_neighbours,
            m=self.m_steps,
            niters=self.niters,
            rademacher=self.rademacher,
            normalized_laplacian=self.normalized_laplacian,
            normalize=self.normalize,
            msid_mode=self.msid_mode,
        )

    def compute(
        self,
        real_features: Union[Tensor, np.ndarray],
        fake_features: Union[Tensor, np.ndarray],
        **kwargs,
    ) -> MetricResult:
        """Compute the Multi-Scale Intrinsic Distance metric.

        Parameters
        ----------
        real_features : Union[Tensor, np.ndarray]
            Array-like of shape (n_samples, n_features) containing features of real samples.
        fake_features : Union[Tensor, np.ndarray]
            Array-like of shape (n_samples, n_features) containing features of fake/generated samples.

        Returns
        -------
        result : MetricResult
            Dataset-level MSID score.
        """
        real_features = to_tensor(real_features)
        fake_features = to_tensor(fake_features)
        score = self._msid.compute_metric(real_features, fake_features)
        return MetricResult(
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": score.detach().item()},
        )


__all__ = ["FrechetDistance", "GeometryScore", "MultiScaleIntrinsicDistance"]



from typing import List, Optional

from .tsfel import TSFEL


class ExtractorFactory:
    default = "tsfel"

    @staticmethod
    def model_from_name(
        name: str,
        domains: Optional[List[str]] = None,
        **kwargs,
    ):
        """Initializes the feature extractor with the given parameters.

        Args:
            name (str): identifier of the extractor to be used.
            device (str): model device. Defaults to "cpu".
        """
        if name == "tsfel":
            return TSFEL(domains, **kwargs)
        else:
            raise ValueError(f"Model {name} not available.")



from pymdma.tabular.measures.synthesis_val.data.similarity import (
    CoherenceScore,
    StatisticalDivergenceScore,
    StatisticalSimScore,
)
# from pymdma.tabular.measures.synthesis_val.data.utility import Utility
from pymdma.tabular.measures.synthesis_val.feature._shared import (
    Authenticity,
    Coverage,
    Density,
    ImprovedPrecision,
    ImprovedRecall,
)
from pymdma.tabular.measures.synthesis_val.feature.privacy import DCRPrivacy, NNDRPrivacy

__all__ = [
    "ImprovedPrecision",
    "ImprovedRecall",
    "Density",
    "Coverage",
    "Authenticity",
    "NNDRPrivacy",
    "DCRPrivacy",
    "StatisticalSimScore",
    "StatisticalDivergenceScore",
    "CoherenceScore",
    # "Utility",
]



from pymdma.general.measures.prdc import Authenticity, Coverage, Density, ImprovedPrecision, ImprovedRecall

__all__ = [
    "ImprovedPrecision",
    "ImprovedRecall",
    "Authenticity",
    "Density",
    "Coverage",
]



import json
from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple, Union

import numpy as np
from loguru import logger
from torch.utils.data import DataLoader

from pymdma.common.definitions import InputLayer
from pymdma.constants import ReferenceType, ValidationDomain

from .data import SimpleDataset
from .models.features import ExtractorFactory
from .utils.processing import batch_downsample_to_largest

SUPPORTED_FILES = {".png", ".jpg", ".jpeg", ".bmp"}  # TODO might want to add others


def no_collate_fn(data):
    # A simple collate_fn for torch dataloaders that prevents the default
    # collate_fn from stacking the images into a single tensor. Used in input validation

    # images | labels | ids
    return [d[0] for d in data], [d[1] for d in data], [d[2] for d in data]


def source_to_list(data_src: Union[Path, List[Path]]) -> List[Path]:
    """Helper function that transforms a data source to a list of paths to each
    file.

    Args:
        data_src (Union[Path, List[Path]]): source directory or list of paths to the images.

    Returns:
        List[Path]: list of paths to the images.
    """
    if isinstance(data_src, list):
        return data_src
    return [im for im in data_src.iterdir() if im.suffix in SUPPORTED_FILES] if data_src else []


def _annotation_from_folder(data_src: Union[Path, None]) -> Path:
    """Helper function that returns the annotation file from a folder."""
    if not data_src:
        return None

    if data_src.is_dir():
        return next(data_src.glob("*.json"), None)
    elif data_src.suffix == ".json":
        return data_src


class ImageInputLayer(InputLayer):
    """Abstraction layer for handling different types of input data with the
    given requirements from the data auditing module.

    Starts by creating the dataloaders for the reference and target
    datasets (if needed), and then send batches of images in the
    required format by the metrics functions.

    Parameters
    ----------
    validation_domain : ValidationDomain
        valition type (input or synthetic)
    reference_type : ReferenceType
        reference type (dataset or single image)
    target_data : Union[Path, List[Path]]
        data root or list of paths to the target dataset
    reference_data : Optional[Union[Path, List[Path]]], optional
        data root or list of paths to the reference dataset, by default None
    annot_file : Optional[Path], optional
        path to the COCO annotation file of the reference dataset, by default None
    repeat_reference : bool, optional
        If True, the reference dataset will be repeated to match the target dataset size, by default False
    batch_size : int, optional
        Batch size to be in the image batch, by default 20
    device : str, optional
        Device to be used in the feature extraction module, by default "cpu"
    features_cache : Optional[Path], optional
        Path to the cache file for the features, by default
    """

    def __init__(
        self,
        validation_domain: ValidationDomain,
        reference_type: ReferenceType,
        target_data: Union[Path, List[Path]],
        reference_data: Optional[Union[Path, List[Path]]] = None,
        annot_file: Optional[Path] = None,
        repeat_reference: bool = False,
        batch_size: int = 20,
        device: str = "cpu",
        features_cache: Optional[Path] = None,
    ) -> None:
        super().__init__()
        self.val_type = validation_domain
        self.reference_type = reference_type
        self.batch_size = batch_size
        self.device = device
        self.features_cache = features_cache

        annot_file = _annotation_from_folder(annot_file)

        self._annotations = None
        if annot_file is not None:
            annot_file = Path(annot_file)
            assert annot_file.exists() and annot_file.suffix == ".json", f"Invalid annotation file {annot_file}"
            self._annotations = json.load(annot_file.open("r"))
        # ids for the images in instance analysis
        # used later for instance level metrics ids
        self.instance_ids = []  # if not self.annotations else [img["id"] for img in self.annotations.dataset["images"]]

        self.transform = np.asarray

        # do not use default collate_fn for input validation (autoconverts images to tensors in the range of 0-1)
        collate_fn = no_collate_fn if self.val_type == ValidationDomain.INPUT else None

        reference_files = source_to_list(reference_data)
        target_files = source_to_list(target_data)
        if self.reference_type != ReferenceType.NONE:
            # repeat reference dataset to match the target dataset size
            # (only in specific cases like input validation with instance reference type)
            if repeat_reference and len(reference_files) < len(target_files):
                logger.warning(
                    "Reference dataset is smaller than the target dataset. Repeating the reference dataset to match the target dataset size.",
                )
                reference_files = (
                    reference_files * (len(target_files) // len(reference_files))
                    + reference_files[: len(target_files) % len(reference_files)]
                )

            # prepare reference dataloader (real/reference images)
            reference_dataset = SimpleDataset(
                file_paths=reference_files,
                transforms=self.transform,
            )
            self.reference_loader = DataLoader(
                reference_dataset,
                batch_size=self.batch_size if self.batch_size > 0 else len(reference_files),
                shuffle=False,
                num_workers=4,
                collate_fn=collate_fn,  # no need to collate images
            )

            assert self.val_type == ValidationDomain.SYNTH or (
                self.val_type == ValidationDomain.INPUT and len(target_files) == len(reference_files)
            ), "Reference and target datasets must have the same size for full reference input validation."

        # prepare target dataloader (evaluation images)
        target_dataset = SimpleDataset(file_paths=target_files, transforms=self.transform)
        self.target_loader = DataLoader(
            target_dataset,
            batch_size=self.batch_size if self.batch_size > 0 else len(target_files),
            shuffle=False,
            num_workers=4,
            collate_fn=collate_fn,  # no need to collate images
        )

        self.instance_ids = [path.stem for path in target_dataset.img_files]

    def __len__(self):
        return len(self.target_loader.dataset)

    @property
    def annotations(self) -> Dict["str", any]:
        assert self._annotations is not None, "Annotations were not provided for this dataset."
        return self._annotations

    @property
    def dataset_files(self) -> List[Path]:
        return self.target_loader.dataset.img_files

    # TODO separate embeddings to support multiple models like in the case of FID
    def get_embeddings(
        self,
        model_name: str,
        model_instances: Optional[Dict[str, callable]] = None,
        offload_model: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Extracts the features from the target and reference datasets using
        the given model.

        Parameters
        ----------
        model_name : str
            name of the model to be used for feature extraction
        model_instances : Optional[Dict[str, callable]], optional
            dictionary of model instances, by default None
        offload_model : bool, optional
            If True, the model will be offloaded to the CPU after the feature extraction, by default False

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            features of the reference and target datasets
        """

        extractor = None
        if model_instances is not None:
            if model_name in model_instances:
                extractor = model_instances[model_name]
            elif model_name == "default" and ExtractorFactory.default in model_instances:
                extractor = model_instances[ExtractorFactory.default]

        if extractor is None:
            offload_model = True  # unload model when finished
            model_name = ExtractorFactory.default if model_name == "default" else model_name
            extractor = ExtractorFactory.model_from_name(model_name) if extractor is None else extractor

        # extractor = model_instance if model_instance is not None else FeatureExtractor(model_name, device=self.device)
        reference_feats, _labels, _reference_ids = extractor._extract_features_dataloader(
            self.reference_loader,
            device=self.device,
        )
        synthetic_feats, _labels, synthetic_ids = extractor._extract_features_dataloader(
            self.target_loader,
            device=self.device,
        )

        if not self.instance_ids:
            self.instance_ids = synthetic_ids.tolist()

        if offload_model:
            extractor = extractor.to("cpu")
            del extractor

        return reference_feats, synthetic_feats

    @property
    def batched_samples(self) -> Generator[Tuple[np.ndarray], None, None]:
        """Returns batches of images from the target dataset in the required
        format."""
        if self.reference_type == ReferenceType.NONE:
            for no_ref_images, _labels, _img_ids in self.target_loader:
                yield (no_ref_images,)
        else:
            ref_iter, sim_iter = iter(self.reference_loader), iter(self.target_loader)
            for _ in range(len(self.reference_loader)):
                ref_imgs, _labels, _img_ids = next(ref_iter)
                sim_imgs, _labels, _img_ids = next(sim_iter)

                ref_imgs, sim_imgs = batch_downsample_to_largest(ref_imgs, sim_imgs)
                yield ref_imgs, sim_imgs



import os
from pathlib import Path
from typing import Callable, List, Optional

import wfdb
from torch.utils.data import Dataset

# TODO support other formats?
SUPPORTED_FILES = {".dat", ".mat", ".csv"}


def _extract_diagnosis(file_path):
    """Extracts diagnosis information from a header file. Only works for this
    specific .hea file structure.

    Parameters
    ----------
    file_path: str
        The path to the header file.

    Returns
    --------
    diagnosis_list: list
        A list of diagnosis codes extracted from the header file.
    """
    with open(file_path) as f:
        lines = f.readlines()
    diagnosis = lines[15].strip().split(": ")[1]
    diagnosis_list = [diagnosis.strip() for diagnosis in diagnosis.split(",")]

    return diagnosis_list


def _extract_fs_dims(file_path):
    """Extracts the sampling frequency and the dimension names of the signal
    from a header file. Only works for this specific .hea file structure.

    Parameters
    ----------
    file_path: str
        The path to the header file.

    Returns
    -------
    fs : int
        Sampling frequency.
    dims: List(str)
        Names of the signal dimensions.
    """
    with open(file_path) as f:
        lines = f.readlines()
    dims = [lines[i].strip().split(" ")[-1] for i in range(1, 13)]
    fs = lines[0].strip().split(" ")[2]

    return int(fs), dims


def _read_sig_file(file_path):
    """Read a signal file from the supported file extensions.

    Parameters:
    -----------
    file_path: Union[str, Path])
        Path to the file.

    Returns
    --------
    dict
        Dictionary containing the data from the .mat file.

    Raises
    ------
    ValueError
        If a file extension different from .mat is found.
    """

    # Check if the file has a .mat extension
    if file_path.suffix in [".mat", ".dat"]:
        directory_path, file_name = os.path.split(file_path)
        file_path = os.path.join(directory_path, file_name.split(".")[0])
        data = wfdb.rdsamp(file_path)[0]
        return data
    else:
        # Raise a ValueError for files with unsupported extensions
        raise AssertionError(f"Unsupported file extension: {Path(file_path).suffix} (file: {file_path})")
        # ? TBD add more supported file types


class SimpleDataset(Dataset):
    def __init__(
        self,
        file_paths: List[Path],
        transforms: Optional[Callable] = None,
        target_transforms: Optional[Callable] = None,
    ) -> None:
        self.sig_files = file_paths
        self.transform = transforms
        self.target_transform = target_transforms

        # Define fs and dims of the dataset
        ex_hea_path = os.path.splitext(self.sig_files[0])[0] + ".hea"
        self.fs, self.dims = _extract_fs_dims(ex_hea_path)

    def __len__(self):
        return len(self.sig_files)

    def __getitem__(self, idx):
        sig_path = Path(self.sig_files[idx])
        signal = _read_sig_file(sig_path)

        # hea_path = os.path.splitext(sig_path)[0] + ".hea"
        label = [None]  # TODO change latter if needed to: label = _extract_diagnosis(hea_path)

        signal_id = sig_path.stem

        if self.transform:
            signal = self.transform(signal)
        if self.target_transform:
            label = self.target_transform(label)
        return signal, label, signal_id



import os
import sys
from pathlib import Path
from loguru import logger
from typing import Dict, Generator, List, Optional, Tuple, Union

import numpy as np
from torch.utils.data import DataLoader

from pymdma.common.definitions import InputLayer
from pymdma.constants import ReferenceType, ValidationDomain

from .data.simple_dataset import SimpleDataset
from .models.features import ExtractorFactory

# Get the absolute path of the parent directory
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

# Add the parent directory to the Python path if it's not already included
if parent_dir not in sys.path:
    sys.path.insert(0, parent_dir)

SUPPORTED_FILES = {".dat", ".mat", ".csv"}  # TODO might want to add others or change


def _custom_collate_fn(batch):
    # Custom collate function to handle varying sizes
    signal, label, signal_id = zip(*batch)
    return list(signal), list(label), list(signal_id)


def _get_data_files_path(data_src: Union[List[str], Path]) -> List[Path]:
    """Get a list of data files from the given source directory or list of file
    paths.

    Parameters
    ----------
    data_src: (Union[List[str], Path])
        A list of file paths or a directory path.

    Returns
    -------
    List[Path]
        A list of data file paths.

    Raises
    ------
    ValueError
        If an invalid item is encountered or if an unsupported file extension is found.
    """
    if isinstance(data_src, list):
        data_files_path = data_src
    else:
        data_files_path = []
        for item in data_src.iterdir():
            if item.is_file():
                if item.suffix in SUPPORTED_FILES:
                    data_files_path.append(item)
                else:
                    logger.warning(f"Skipping unsupported file extension: {item.suffix} (file: {item})")
            elif item.is_dir():
                item = Path(item)
                # Recursively search for data files in subdirectories
                for sig_file in item.iterdir():
                    if sig_file.is_file() and sig_file.suffix in SUPPORTED_FILES:
                        data_files_path.append(sig_file)
                    else:
                        logger.warning(f"Skipping unsupported file extension: {sig_file.suffix} (file: {sig_file})")
            else:
                raise AssertionError(f"Invalid item encountered: {item}")

    return data_files_path


class TimeSeriesInputLayer(InputLayer):
    """Abstraction layer for handling different types of input data with the
    given requirements from the data auditing module."""

    def __init__(
        self,
        validation_domain: ValidationDomain,
        reference_type: ReferenceType,
        target_data: Union[Path, List[Path]],
        reference_data: Optional[Union[Path, List[Path]]] = None,
        batch_size: int = 64,
        device: str = "cpu",
        **kwargs,
    ) -> None:
        """Initializes the input layer with the given parameters.

        Parameters
        ----------
        validation_domain : ValidationDomain
            Valition type (input or synthetic).
        reference_type: ReferenceType
            Reference type (dataset or single signal).
        data_src1 : Path
            Data root or list of paths to the reference dataset in synthetic evaluation or the dataset in input validation.
        data_src2: (Path, optional)
            Data root or list of paths to the synthetic dataset in synthetic ealuation or another dataset used for full reference input validation. Defaults to None.
        batch_size: int
            Batch size to be used in the feature extraction module or the number of signals returned in the input validation. Defaults to 64.
        device : str
            Device to be used in the feature extraction module. Defaults to "cpu".
        feature_extractor_name : str
            Feature extractor name. Defaults to tsfel.
        """
        super().__init__()
        self.val_type = validation_domain
        self.reference_type = reference_type
        self.batch_size = batch_size
        self.device = device

        # ids for the signals in instance analysis
        # used later for instance level metrics ids
        self.instance_ids = []

        # prepare feature extractor for synthetic evaluation
        self.transform = np.asarray  # default transform to np.array

        collate_fn = _custom_collate_fn

        target_files = _get_data_files_path(target_data)

        # prepare reference dataloader (original/reference signals)
        # will also be used for input validation
        if reference_type != ReferenceType.NONE:
            reference_files = _get_data_files_path(reference_data)
            reference_dataset = SimpleDataset(file_paths=reference_files, transforms=self.transform)
            self.reference_loader = DataLoader(
                reference_dataset,
                batch_size=self.batch_size if self.batch_size > 0 else len(reference_files),
                shuffle=False,
                num_workers=4,
                collate_fn=collate_fn,
            )

        # prepare target dataloader (synthetic/similar signals)
        # Will be used for synthetic evaluation or full reference input validation
        target_dataset = SimpleDataset(file_paths=target_files, transforms=self.transform)
        self.target_loader = DataLoader(
            target_dataset,
            batch_size=self.batch_size if self.batch_size > 0 else len(target_dataset),
            shuffle=False,
            num_workers=4,
            collate_fn=collate_fn,
        )

        # TODO review this for modality
        # Client input validation
        if self.val_type == ValidationDomain.SYNTH:
            assert (
                len(target_files) > 4
            ), f"Synthetic datasets must have at least 5 signals for synthetic evaluation. Found {len(target_files)}."
        elif self.val_type == ValidationDomain.INPUT and reference_type != ReferenceType.NONE:
            assert len(reference_dataset) == len(
                target_dataset,
            ), "Reference and target datasets must have the same size for full reference input validation."

        self.instance_ids = [path.stem for path in target_dataset.sig_files]
        # if feature_extractor_name == "tsfel":
        #     self.target_loader.fs = target_dataset.fs
        #     self.target_loader.dims = target_dataset.dims

    def __len__(self):
        return len(self.target_loader.dataset)

    def get_embeddings(
        self,
        model_name: str,
        model_instances: Optional[Dict[str, callable]] = None,
        offload_model: bool = False,
    ):
        extractor = None
        if model_instances is not None:
            if model_name in model_instances:
                extractor = model_instances[model_name]
            elif model_name == "default" and ExtractorFactory.default in model_instances:
                extractor = model_instances[ExtractorFactory.default]
        if extractor is None:
            model_name = ExtractorFactory.default if model_name == "default" else model_name
            extractor = ExtractorFactory.model_from_name(model_name) if extractor is None else extractor

        reference_features, _labels, _ = extractor._extract_features_dataloader(
            self.reference_loader,
            self.reference_loader.dataset.fs,
            self.reference_loader.dataset.dims,
        )
        synth_features, _labels, self.instance_ids = extractor._extract_features_dataloader(
            self.target_loader,
            self.target_loader.dataset.fs,
            self.target_loader.dataset.dims,
        )

        if offload_model and model_name != "tsfel":
            extractor._model = extractor._model.to("cpu")
            del extractor
        return reference_features, synth_features

    @property
    def batched_samples(self) -> Generator[Tuple[np.ndarray], None, None]:
        # only reference signals for no reference metrics
        if self.reference_type == ReferenceType.NONE:
            for no_ref_signals, _labels, _sig_ids in self.target_loader:
                # save signal ids for later
                # self.instance_ids.extend(sig_ids)
                yield (no_ref_signals,)
        else:  # full reference
            # iterate through both dataloaders and yield batches of signals
            ref_iter, sim_iter = iter(self.reference_loader), iter(self.target_loader)
            for _ in range(len(self.reference_loader)):
                ref_sigs, _, _ = next(ref_iter)
                sim_sigs, _, _sig_ids = next(sim_iter)

                # save signal ids for later
                # self.instance_ids.extend(list(sig_ids))

                yield ref_sigs, sim_sigs

    def get_full_samples(self):
        # only reference signals for no reference metrics
        if self.reference_type == ReferenceType.NONE:
            full_no_ref_signals = []
            for no_ref_signals, _labels, _sig_ids in self.target_loader:
                full_no_ref_signals.extend(no_ref_signals)
            return np.array(full_no_ref_signals)
        else:  # full reference
            # iterate through both dataloaders and return all signals
            full_ref_sigs = []
            full_sim_sigs = []
            ref_iter = iter(self.reference_loader)
            sim_iter = iter(self.target_loader)
            for _ in range(len(self.reference_loader)):
                ref_sigs, _, _ = next(ref_iter)
                sim_sigs, _, _sig_ids = next(sim_iter)
                full_ref_sigs.extend(ref_sigs)
                full_sim_sigs.extend(sim_sigs)

            return np.array(full_ref_sigs), np.array(full_sim_sigs)



from typing import List

import numpy as np
import pandas as pd


def is_float(
    string: str,
):
    """
    Check if the provided string can be converted to a float.

    Parameters
    ----------
    string : str
        The string to check for float conversion.

    Returns
    -------
    bool
        True if the string can be converted to float, False otherwise.
    """
    
    try:
        float(string)
        return True
    except ValueError:
        return False


def is_categorical(
    data: np.ndarray,
    thresh: int = 10,
):
    """
    Determine whether the provided data is categorical based on the ratio of unique values.

    Parameters
    ----------
    data : np.ndarray
        The data to be evaluated for categorical nature.
    thresh : float, optional
        The threshold ratio of non-unique values to unique values for categorization. 
        Default is 0.4.

    Returns
    -------
    bool
        True if the data is categorical based on the threshold, False otherwise.
    """
    
    # Number of samples
    len_data = len(data)

    # Number of unique samples
    len_unq = len(np.unique(data))

    # Ratio
    ratio = (len_data - len_unq) / (len_data + len_unq)

    return len_unq < thresh


def is_numeric(
    data: np.ndarray,
):
    """
    Check if all values in the provided data can be converted to numeric values.

    Parameters
    ----------
    data : np.ndarray
        The data to check for numeric conversion.

    Returns
    -------
    bool
        True if the data is numeric, False otherwise.
    """
    
    try:
        # check if it is conversible to float
        _ = np.array(data).astype(float)
        is_num = True
    except Exception as e:
        is_num = False
    return is_num

def is_integer(
    data: np.ndarray,
):
    """
    Check if all values in the provided data can be converted to integer values.

    Parameters
    ----------
    data : np.ndarray
        The data to check for integer conversion.

    Returns
    -------
    bool
        True if the data is integer, False otherwise.
    """
    try:
        # for numerical arrays
        is_int = (np.array(data, int) == data).all()
    except Exception as e:
        # for non-numerical arrays
        is_int = False

    return is_int

def replace_missings_df(
    data: pd.DataFrame,
    missing_values: List = [999, -1, "NaN", "nan"],
    **kwargs,
):
    """
    Replace specified missing values in a pandas DataFrame with NaNs.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame where missing values need to be replaced.
    missing_values : list, optional
        A list of values to treat as missing (default is [999, -1, "NaN", "nan"]).

    Returns
    -------
    pd.DataFrame
        DataFrame with specified missing values replaced by NaN.
    """
    
    return data.replace(missing_values, np.nan)


def replace_missings_np(
    data: np.ndarray,
    missing_values: List = [999, -1, "NaN", "nan"],
    **kwargs,
):
    """
    Replace specified missing values in a NumPy array with NaNs.

    Parameters
    ----------
    data : np.ndarray
        The array where missing values need to be replaced.
    missing_values : list, optional
        A list of values to treat as missing (default is [999, -1, "NaN", "nan"]).

    Returns
    -------
    np.ndarray
        A new array with specified missing values replaced by NaN.
    """
    
    # copy of the original vector
    aux_data = data.copy()

    # replace all the instances of missing values with NaN tag
    aux_data[np.isin(aux_data, missing_values)] = np.nan

    return aux_data


def auto_metadata_generator(
    data: np.ndarray,
    cols: list,
    **kwargs,
):
    """
    Automatically generate metadata about the columns of a dataset, identifying
    whether each column is categorical or numerical.

    Parameters
    ----------
    data : np.ndarray
        The data for which metadata will be generated.
    cols : list of str
        A list of column names corresponding to the data columns.
    **kwargs : optional
        Additional keyword arguments for handling missing values or other preprocessing.

    Returns
    -------
    dict
        A dictionary containing metadata for each column, including type (categorical
        or continuous) and data type (numerical or categorical).
    """
    
    vtype_d = {}

    # encode missing values
    data_ = replace_missings_np(data, **kwargs)

    # Loop over columns
    for idx, col_ in enumerate(cols):
        # auxiliary vector
        scores_d = {}

        # dtypes
        is_num = is_numeric(data_[:, idx])
        is_categ = is_categorical(data_[:, idx])
        is_int = is_integer(data_[:, idx])

        if is_categ or not is_num:
            # unique values
            opt_unq = np.unique(data_[:, idx]).astype(float if is_num else str)

            # remove NaN values
            opt_unq = opt_unq[~pd.isnull(opt_unq)]

            # encode
            opt_map = {v: k for k, v in enumerate(opt_unq, 1)}

            # categorical check
            is_categ = int((len(opt_unq)/len(data)) < 0.03)  # needs to be less than 3%

            # assign
            scores_d["type"] = {"tag": "discrete", "is_categ": is_categ, "opt": opt_map}
        else:
            # assign
            scores_d["type"] = {"tag": "continuous", "is_categ": 0, "opt": {}}

        # data type 
        if is_num:
            scores_d["dtype"] = "numerical" if not is_int else "integer"
        else:
            scores_d["dtype"] = "string"

        # assign
        vtype_d[col_] = scores_d

    return vtype_d


def get_dtypes_from_type_map(vtype_map: dict):
    """
    Automatically generate metadata about the columns of a dataset, identifying
    whether each column is categorical or numerical.

    Parameters
    ----------
    data : np.ndarray
        The data for which metadata will be generated.
    cols : list of str
        A list of column names corresponding to the data columns.
    **kwargs : optional
        Additional keyword arguments for handling missing values or other preprocessing.

    Returns
    -------
    dict
        A dictionary containing metadata for each column, including type (categorical
        or continuous) and data type (numerical or categorical).
    """
    
    aux_map = {"numerical": "float", "string": "str", "integer": "int"}

    # define variable mappers
    opt_map, dtype_map = {}, {}

    # get data type mapper per column
    for col, vtype in vtype_map.items():
        # column option range
        if vtype["type"]["tag"] != "numerical":
            opt_map[col] = vtype["type"].get("opt")
        else:
            opt_map[col] = {}

        # dtype options
        dtype_map[col] = aux_map.get(vtype["dtype"], "str")

    return dtype_map, opt_map



# fetch common metrics from the general module
from pymdma.image.measures.synthesis_val.feature._shared import (
    Authenticity,
    Coverage,
    Density,
    FrechetDistance,
    GeometryScore,
    ImprovedPrecision,
    ImprovedRecall,
    MultiScaleIntrinsicDistance,
    PrecisionRecallDistribution,
)
from pymdma.image.measures.synthesis_val.feature.giqa import GIQA

__all__ = [
    "GIQA",
    "ImprovedPrecision",
    "ImprovedRecall",
    "Density",
    "Coverage",
    "Authenticity",
    "FrechetDistance",
    "GeometryScore",
    "MultiScaleIntrinsicDistance",
    "PrecisionRecallDistribution",
]



import numpy as np

from pymdma.common.definitions import FeatureMetric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

from ..utils.util import cluster_into_bins


def _compute_prd(eval_dist, ref_dist, num_angles: int = 1001, epsilon: float = 1e-10):
    """Computes the PRD curve for discrete distributions.

    This function computes the PRD curve for the discrete distribution eval_dist
    with respect to the reference distribution ref_dist. This implements the
    algorithm in [arxiv.org/abs/1806.2281349]. The PRD will be computed for an
    equiangular grid of num_angles values between [0, pi/2].

    Parameters
    ----------
    eval_dist : np.ndarray or list of float
        1D array or list of floats with the probabilities of the different states
        under the distribution to be evaluated.
    ref_dist : np.ndarray or list of float
        1D array or list of floats with the probabilities of the different states
        under the reference distribution.
    num_angles : int, optional
        Number of angles for which to compute PRD. Must be in [3, 1e6].
        The default value is 1001.
    epsilon : float, optional
        Angle for PRD computation in the edge cases 0 and pi/2. The PRD
        will be computed for epsilon and pi/2-epsilon, respectively.
        The default value is 1e-10.

    Returns
    -------
    precision : numpy.ndarray, shape (num_angles,)
        Precision for the different ratios.
    recall : numpy.ndarray, shape (num_angles,)
        Recall for the different ratios.

    Raises
    ------
    AssertionError
        If `epsilon` is not in the range (0, 0.1].
    AssertionError
        If `num_angles` is not in the range [3, 1e6].


    References
    ---------
    Sajjadi, Mehdi SM, et al. Assessing generative models via precision and recall (2018).
    https://proceedings.neurips.cc/paper_files/paper/2018/file/f7696a9b362ac5a51c3dc8f098b73923-Paper.pdf

    Code adapted from:
    https://github.com/vanderschaarlab/evaluating-generative-models/blob/main/metrics/prd_score.py
    """

    assert epsilon > 0 and epsilon <= 0.1, "epsilon must be in (0, 0.1]."
    assert num_angles >= 3 and num_angles <= 1e6, "num_angles must be in [3, 1e6]."

    # Compute slopes for linearly spaced angles between [0, pi/2]
    angles = np.linspace(epsilon, np.pi / 2 - epsilon, num=num_angles)
    slopes = np.tan(angles)

    # Broadcast slopes so that second dimension will be states of the distribution
    slopes_2d = np.expand_dims(slopes, 1)

    # Broadcast distributions so that first dimension represents the angles
    ref_dist_2d = np.expand_dims(ref_dist, 0)
    eval_dist_2d = np.expand_dims(eval_dist, 0)

    # Compute precision and recall for all angles in one step via broadcasting
    precision = np.minimum(ref_dist_2d * slopes_2d, eval_dist_2d).sum(axis=1)
    recall = precision / slopes

    # handle numerical instabilities leaing to precision/recall just above 1
    precision = np.clip(precision, 0, 1)
    recall = np.clip(recall, 0, 1)
    return precision, recall


def _prd_to_f_beta(precision, recall, beta=1, epsilon=1e-10):
    """Computes F_beta scores for the given precision/recall values.

    The F_beta scores for all precision/recall pairs will be computed and
    returned.

    For precision p and recall r, the F_beta score is defined as:
    F_beta = (1 + beta^2) * (p * r) / ((beta^2 * p) + r)

    Parameters
    ----------
    precision : np.ndarray
        1D NumPy array of precision values in [0, 1].
    recall : np.ndarray
        1D NumPy array of recall values in [0, 1].
    beta : float, optional
        Beta parameter. Must be positive. The default value is 1.
    epsilon : float, optional
        Small constant to avoid numerical instability caused by division by 0
        when precision and recall are close to zero.

    Returns
    -------
    np.ndarray
        NumPy array of same shape as precision and recall with the F_beta scores
        for each pair of precision/recall.

    Raises
    ------
    AssertionError
        If `beta` is not positive.

    References
    ---------
    Sajjadi, Mehdi SM, et al. Assessing generative models via precision and recall (2018).
    https://proceedings.neurips.cc/paper_files/paper/2018/file/f7696a9b362ac5a51c3dc8f098b73923-Paper.pdf

    Code adapted from:
    https://github.com/vanderschaarlab/evaluating-generative-models/blob/main/metrics/prd_score.py
    """
    assert beta > 0, "Given parameter beta must be positive."

    return (1 + beta**2) * (precision * recall) / ((beta**2 * precision) + recall + epsilon)


class PrecisionRecallDistribution(FeatureMetric):
    """Computes PRD data from sample embeddings and the maximum F_beta scores
    for the given precision/recall values.

    The points from both distributions are mixed and then clustered. This leads
    to a pair of histograms of discrete distributions over the cluster centers
    on which the PRD algorithm is executed.

    For PRD, it is recommended that number of points in eval_data and ref_data are equal since
    unbalanced distributions bias the clustering towards the larger dataset. The
    check for this condition can be performed by setting the enforce_balance flag to True
    (recommended).

    Regarding the maximum F_beta scores, the maximum F_beta score over all pairs of precision/recall
    values is useful to compress a PRD plot into a single value which correlate with recall.
    Whereas, the max_f_beta_inv score over all pairs of precision/recall values compresses
    the PRD plot into a single value that correlates with precision.

    **Objective**: Fidelity, Diversity

    Parameters
    ----------
    num_clusters: int, optional
        Number of cluster centers to fit. The default value is 2.
    num_angles: int, optional
        Number of angles for which to compute PRD. Must be in [3, 1e6]. The default value is 1001.
    num_runs: int, optional
        Number of independent runs over which to average the PRD data. The default value is 10.
    beta: int, optional
        Beta parameter for F_beta score. Must be positive. The default value is 8.
    epsilon: float, optional
        Small constant to avoid numerical instability caused by division
        by 0 when precision and recall are close to zero. The default value is 1e-10.
    compute_stats: bool, optional
        If True, F_beta scores for all precision/recall pairs will be computed.
        If False, F_beta scores computation is skipped. Default is True.

    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unused).

    References
    ---------
    Sajjadi, Mehdi SM, et al. Assessing generative models via precision and recall (2018).
    https://proceedings.neurips.cc/paper_files/paper/2018/file/f7696a9b362ac5a51c3dc8f098b73923-Paper.pdf

    Code adapted from:
    https://github.com/vanderschaarlab/evaluating-generative-models/blob/main/metrics/prd_score.py


    Examples
    --------
    >>> prd = PrecisionRecallDistribution()
    >>> x_feats = np.random.rand(64, 48)
    >>> y_feats = np.random.rand(64, 48)
    >>> result: MetricResult = prd.compute(x_feats, y_feats)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    def __init__(
        self,
        num_clusters: int = 2,
        num_angles: int = 1001,
        num_runs: int = 10,
        epsilon: float = 1e-10,
        beta: int = 8,
        compute_stats: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_clusters = num_clusters
        self.num_angles = num_angles
        self.num_runs = num_runs
        self.epsilon = epsilon
        self.compute_stats = compute_stats

        assert beta > 0, "Given parameter beta must be positive."
        self.beta = beta

    def _prd_from_embedding(self, target: np.ndarray, reference: np.ndarray):
        eval_data = np.array(target, dtype=np.float64)
        ref_data = np.array(reference, dtype=np.float64)
        precisions = []
        recalls = []
        for _ in range(self.num_runs):
            eval_dist, ref_dist = cluster_into_bins(eval_data, ref_data, self.num_clusters)
            precision, recall = _compute_prd(eval_dist, ref_dist, self.num_angles, self.epsilon)
            precisions.append(precision)
            recalls.append(recall)
        precision = np.mean(precisions, axis=0)
        recall = np.mean(recalls, axis=0)
        return precision, recall

    def compute(self, real_features: np.ndarray, fake_features: np.ndarray, **kwargs) -> MetricResult:
        """Computes PRD data from sample embeddings. Using the PRD, the maximum
        F_beta and F_beta_inv score are also computed. These scores for the
        given precision/recall values correlate, respectively, with recall and
        precision.

        Returns
        -------
        MetricResult
            Instance-level PRD values.
            Dataset-leve max_f_beta and max_f_beta_inv scores.
        """
        warning = None
        if len(fake_features) != len(real_features):
            warning = (
                "The number of points in eval_data %d is not equal to the number of "
                "points in ref_data %d. To disable this exception, set enforce_balance "
                "to False (not recommended)." % (len(fake_features), len(real_features))
            )

        precision, recall = self._prd_from_embedding(fake_features, real_features)

        stats = None
        if self.compute_stats:
            max_f_beta = np.max(_prd_to_f_beta(precision, recall, self.beta, self.epsilon))
            max_f_beta_inv = np.max(_prd_to_f_beta(recall, precision, 1 / self.beta, self.epsilon))
            stats = {
                "max_f_beta": max_f_beta,
                "max_f_beta_inv": max_f_beta_inv,
            }

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.KEY_ARRAY,
                "subtype": "float",
                "value": {
                    "precision_values": precision,
                    "recall_values": recall,
                },
                "plot_params": {
                    "x_label": "Recall",
                    "y_label": "Precision",
                    "kind": "line",
                    "x_key": "recall_values",
                    "y_key": "precision_values",
                },
                "stats": stats,
            },
            errors=[warning] if warning else None,
        )


__all__ = ["PrecisionRecallDistribution"]



from typing import Literal, Tuple, Union

import cv2
import numpy as np
import torch
from PIL import Image, ImageEnhance
from piq.brisque import brisque
from piq.clip_iqa import CLIPIQA as _clip_iqa

from pymdma.common.definitions import Metric
from pymdma.common.output import DistributionResult, MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

from ....utils.processing import image_resize


class Tenengrad(Metric):
    """Computes Tenengrad score for an image. Sharpness measure based on the
    gradient magnitude.

    **Objective**: Sharpness

    Parameters
    ----------
    kernel_size : int, optional, default=3
        Size of the Sobel kernel.
    threshold : int, optional, default=0
        Threshold for valid gradient pixels. Used to supress noise, smooth the curve and impose sensitivity.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Groen et al., A Comparison of Different Focus Functions for Use in Autofocus Algorithms (1984)
    https://onlinelibrary.wiley.com/doi/pdf/10.1002/cyto.990060202

    More information on Tenengrad:
    Her et al., Research of Image Sharpness Assessment Algorithm for Autofocus (2019)
    https://ieeexplore.ieee.org/abstract/document/8980980

    Examples
    --------
    >>> tenengrad = Tenengrad()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = tenengrad.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        kernel_size: int = 3,
        threshold: int = 0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.threshold = threshold
        self.ksize = kernel_size

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes Tenengrad score for a list of images.

        Parameters
        ----------
        imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.

        Returns
        -------
        result: MetricResult
            Tenengrad score for each image.
        """
        scores = []
        for img in imgs:
            if len(img.shape) == 3:
                img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

            gradient_x = cv2.Sobel(img, cv2.CV_64F, 1, 0, ksize=self.ksize)
            gradient_y = cv2.Sobel(img, cv2.CV_64F, 0, 1, ksize=self.ksize)

            x_y_gradients = np.sqrt(gradient_x**2 + gradient_y**2)

            x_y_gradients[x_y_gradients < self.threshold] = 0

            scores.append(np.mean(x_y_gradients))

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": scores},
        )


class TenengradRelative(Metric):
    """Computes Tenengrad score for an image in relation to a blurred instance
    of itself. Sharpness measure based on the gradient magnitude.

    **Objective**: Sharpness

    Parameters
    ----------
    kernel_size : int, optional, default=3
        Size of the Sobel kernel.
    threshold : int, optional, default=0
        Threshold for valid gradient pixels. Used to supress noise, smooth the curve and impose sensitivity.
    criteria : str, optional, default="ratio"
        Criteria for relative tenengrad score. Can be `ratio` or `diff`.
    blur_factor : float, optional, default=0.0
        Degree of blurring to apply to the image. The lower the value, the more blurred the image.
        Must be in the range [0., 1.].
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Groen et al., A Comparison of Different Focus Functions for Use in Autofocus Algorithms (1984)
    https://onlinelibrary.wiley.com/doi/pdf/10.1002/cyto.990060202

    More information on Tenengrad:
    Her et al., Research of Image Sharpness Assessment Algorithm for Autofocus (2019)
    https://ieeexplore.ieee.org/abstract/document/8980980

    Examples
    --------
    >>> tenengrad = TenengradRelative()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = tenengrad.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        kernel_size: int = 3,
        threshold: int = 0,
        criteria: Literal["ratio", "diff"] = "ratio",
        blur_factor: float = 0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.ksize = kernel_size
        self.trheshold = threshold
        assert criteria in ["ratio", "diff"], f"Unsupported criteria for relative tenengrad: {criteria}"
        self.criteria = criteria

        assert 0.0 <= blur_factor <= 1.0, "Blur factor must be in the range [0., 1.]"
        self.blur_factor = blur_factor

        self.tenengrad = Tenengrad(ksize=kernel_size, threshold=threshold)

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes TenengradRelative score for a list of images.

        Parameters
        ----------
        imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.

        Returns
        -------
        result: MetricResult
            Tenengrad score for each image.
        """
        relative_metrics = []
        for img in imgs:
            img_blur = np.asarray(ImageEnhance.Sharpness(Image.fromarray(img)).enhance(self.blur_factor))

            tenengrad_metric_img = self.tenengrad.compute([img]).instance_level.value[0]
            tenengrad_metric_img_blur = self.tenengrad.compute([img_blur]).instance_level.value[0]

            relative_metrics.append(
                (
                    tenengrad_metric_img_blur / tenengrad_metric_img
                    if self.criteria == "ratio"
                    else tenengrad_metric_img_blur - tenengrad_metric_img
                ),
            )

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": relative_metrics},
        )


class EME(Metric):
    """
    Computes the Measure of Enhancement (EME) score.
    Quantifies the enhancement of an image by measuring the contrast ratio of the image.
    Adapted from: https://www.researchgate.net/publication/244268659_A_New_Measure_of_Image_Enhancement

    **Objective**: Contrast

    Parameters
    ----------
    blocks_size : tuple of int, optional, default=(100, 100)
        Size of the blocks to divide the image into.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Agaian et al., A New Measure of Image Enhancement (2000).
    https://www.researchgate.net/publication/244268659_A_New_Measure_of_Image_Enhancement

    Examples
    --------
    >>> eme = EME()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = eme.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        blocks_size: Tuple[int, int] = (100, 100),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.block_size = blocks_size

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes the Measure of Enhancement (EME) score for a list of
        images.

        Parameters
        ----------
        imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.

        Returns
        -------
        result: MetricResult
            EME score for each image.
        """
        emes = []
        for img in imgs:
            if len(img.shape) == 3:
                img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
            img_width = img.shape[0]
            img_height = img.shape[1]
            n_steps_width = int(img_width / self.block_size[0])
            n_steps_height = int(img_height / self.block_size[1])
            total_contrast_ratio = 0
            total_blocks = 0
            for w_step in range(0, n_steps_width):
                w_cords = (w_step * self.block_size[0], (w_step + 1) * self.block_size[0])
                for h_step in range(1, n_steps_height):
                    h_cords = (h_step * self.block_size[1], (h_step + 1) * self.block_size[1])
                    cur_block = img[w_cords[0] : w_cords[1], h_cords[0] : h_cords[1]]
                    if np.min(cur_block) < 0.1:
                        min_val = 0.1
                    else:
                        min_val = np.min(cur_block)
                    if np.max(cur_block) < 0.1:
                        max_val = 0.1
                    else:
                        max_val = np.max(cur_block)
                    contrast_ratio_block = 20 * np.log(max_val / min_val)
                    total_contrast_ratio += contrast_ratio_block
                    total_blocks += 1

            emes.append(total_contrast_ratio / total_blocks if total_blocks > 0 else 0)

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": emes},
        )


# TODO documentation
class ExposureBrightness(Metric):
    """Computes Exposure and Brightness level Metric. Values higher than 1
    indicate overexposure, while values closer to 0 indicate underexposure.

    **Objective**: Exposure and Brightness

    Parameters
    ----------
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    Examples
    --------
    >>> exposure_brightness = ExposureBrightness()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = exposure_brightness.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes exposure level for a list of images.

        Parameters
        ----------
        imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.

        Returns
        -------
        result: MetricResult
            Exposure score for each image.
        """
        exposure_levels = []
        for img in imgs:
            if len(img.shape) == 3:
                img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

            mean, std = cv2.meanStdDev(img)
            exposure_levels.append((mean[0][0] + 2 * std[0][0]) / 255)

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": exposure_levels},
        )


class Brightness(Metric):
    """Computes brightness level of an image.

    **Objective**: Brightness

    Parameters
    ----------
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Darel Rex Finley, HSP Color Model — Alternative to HSV (HSB) and HSL (2006).
    http://alienryderflex.com/hsp.html

    Marian Stefanescu, Measuring and enhancing image quality attributes (2021).
    https://towardsdatascience.com/measuring-enhancing-image-quality-attributes-234b0f250e10

    Examples
    --------
    >>> brightness = Brightness()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = brightness.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes brightness level for a list of images.

        Parameters
        ----------
        imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.

        Returns
        -------
        result: MetricResult
            Brightness score for each image.
        """
        scores = []
        for img in imgs:
            assert img.shape[-1] == 3, "Image should be in RGB format"
            nr_of_pixels = len(img) * len(img[0])

            img = img.astype(np.float32)
            r_vals = 0.299 * img[:, :, 0] ** 2
            g_vals = 0.587 * img[:, :, 1] ** 2
            b_vals = 0.114 * img[:, :, 2] ** 2
            total_brightness = np.sqrt(r_vals + g_vals + b_vals).sum()

            scores.append(total_brightness / nr_of_pixels)

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": scores},
        )


class Colorfulness(Metric):
    """Computes colorfulness level of an image.

    **Objective**: Colorfulness

    Parameters
    ----------
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Hasler et al., Measuring colourfulness in natural images (2003).
    https://infoscience.epfl.ch/server/api/core/bitstreams/77f5adab-e825-4995-92db-c9ff4cd8bf5a/content

    Code was adapted from:
    Adrian Rosebrock, Computing image “colorfulness” with OpenCV and Python (2017).
    https://pyimagesearch.com/2017/06/05/computing-image-colorfulness-with-opencv-and-python/

    Examples
    --------
    >>> colorfulness = Colorfulness()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = colorfulness.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes colorfulness level of list of images.

        Parameters
        ----------
        imgs : (N, H, W, C) ndarray
            List of arrays representing RGB image of shape (H, W, C).

        Returns
        -------
        result: MetricResult
            Colorfulness score for each image.
        """
        scores = []
        for img in imgs:
            assert len(img.shape) == 3, "Image should be in RGB format"

            img = img.astype(np.float32)
            (r_channel, g_channel, b_channel) = img[:, :, 0], img[:, :, 1], img[:, :, 2]

            rg = np.absolute(r_channel - g_channel)
            yb = np.absolute(0.5 * (r_channel + g_channel) - b_channel)

            (rb_mean, rb_std) = (np.mean(rg), np.std(rg))
            (yb_mean, yb_std) = (np.mean(yb), np.std(yb))

            std_root = np.sqrt((rb_std**2) + (yb_std**2))
            mean_root = np.sqrt((rb_mean**2) + (yb_mean**2))

            scores.append(std_root + (0.3 * mean_root))
        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": scores},
        )


class CLIPIQA(Metric):
    """Compute the CLIP-based IQA score. Wrapper of the PIQ CLIP-IQA metric.
    Evaluates perceptual quality of an image using a CLIP model.

    **Objective**: General Image Quality

    Parameters
    ----------
    img_size : {tuple of int, int}, optional, default=(512, 512)
        Size for each image. If a single integer is provided, the image will be thumbnail resized.
        In thumbnail resizing, the aspect ratio is maintained and batch calculation is not allowed (slower computation).
    interpolation : int, optional, default=cv2.INTER_LINEAR
        Interpolation method for resizing.
    data_range : float, optional, default=255
        The range of the data. By default, it is assumed to be [0, 255].
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Wang et al., Exploring CLIP for Assessing the Look and Feel of Images (2022).
    https://arxiv.org/abs/2207.12396

    This is a wrapper class for the implementation in:
    piq, PyTorch Image Quality: Metrics for Image Quality Assessment.
    https://github.com/photosynthesis-team/piq

    Examples
    --------
    >>> clip_iqa = CLIPIQA()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = clip_iqa.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        img_size: Union[Tuple[int, int], int] = (512, 512),
        interpolation: int = cv2.INTER_LINEAR,
        data_range: float = 255,
        device: str = "cpu",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.img_size = img_size
        self.interpolation = interpolation
        self._clip = _clip_iqa(data_range=data_range).to(device)
        self.device = device

        if isinstance(img_size, tuple):
            self._height, self._width = img_size, img_size
            self._batch_calculation = True
        else:
            # thumbnail resize (different image sizes)
            self._height, self._width = img_size, None
            self._batch_calculation = False

        self._height, self._width = img_size if isinstance(img_size, tuple) else (img_size, None)

    def _process_image(self, img: np.ndarray) -> np.ndarray:
        """Resize image to the required size and convert to tensor."""
        img = image_resize(img, height=self._height, width=self._width, inter=self.interpolation)
        return torch.tensor(img).permute(2, 0, 1)

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes CLIPIQA level of a list of images.

        Parameters
        ----------
        imgs : (N, H, W, C) ndarray
            List of arrays representing RGB image of shape (H, W, C).

        Returns
        -------
        result: MetricResult
            CLIPIQA score for each image.
        """
        imgs = [self._process_image(img) for img in imgs]

        if self._batch_calculation:
            imgs = torch.stack(imgs).to(self.device)
            scores = self._clip(imgs).detach().cpu().squeeze(1).tolist()
        else:
            scores = [self._clip(img.unsqueeze(0)).detach().cpu().item() for img in imgs]

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": scores},
        )


class BRISQUE(Metric):
    """Computes  Blind/referenceless Image Spatial Quality Evaluator (BRISQUE)
    score. Wrapper of the PIQ BRISQUE metric implementation.

    **Objective**: General Image Quality

    Parameters
    ----------
    kernel_size : int, optional, default=7
        The size of the Gaussian kernel.
    kernel_sigma : float, optional, default=7 / 6
        The standard deviation of the Gaussian kernel.
    data_range : float, optional, default=255
        The range of the data. By default, it is assumed to be [0, 255].
    device : str, optional, default="cpu"
        Device to run the computation.
    same_size : bool, optional, default=False
        If True, all provided images must have the same size (faster computation).
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Mittal et al., No-Reference Image Quality Assessment in the Spatial Domain (2012).
    https://live.ece.utexas.edu/publications/2012/TIP%20BRISQUE.pdf

    This is a wrapper class for the implementation in:
    piq, PyTorch Image Quality: Metrics for Image Quality Assessment.
    https://github.com/photosynthesis-team/piq

    Examples
    --------
    >>> brisque = BRISQUE()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = brisque.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        kernel_size: int = 7,
        kernel_sigma: float = 7 / 6,
        data_range: float = 255,
        device: str = "cpu",
        same_size: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.kernel_size = kernel_size
        self.kernel_sigma = kernel_sigma
        self.data_range = data_range
        self._brisque = brisque
        self._batch_calculation = same_size
        self.device = device

    def _process_image(self, img: np.ndarray) -> np.ndarray:
        """Resize image to the required size and convert to tensor."""
        return torch.tensor(img).permute(2, 0, 1)

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes BRISQUE.

        Parameters
        ----------
        imgs : (N, H, W, C) ndarray
            List of arrays representing RGB image of shape (H, W, C).

        Returns
        -------
        result: MetricResult
            BRISQUE score for each image.
        """
        imgs = [self._process_image(img) for img in imgs]

        if self._batch_calculation:
            imgs = torch.stack(imgs).to(self.device)
            scores = (
                self._brisque(
                    imgs,
                    kernel_size=self.kernel_size,
                    kernel_sigma=self.kernel_sigma,
                    data_range=self.data_range,
                    reduction="none",
                )
                .detach()
                .cpu()
                .tolist()
            )
        else:
            scores = [
                self._brisque(
                    img.unsqueeze(0),
                    kernel_size=self.kernel_size,
                    kernel_sigma=self.kernel_sigma,
                    data_range=self.data_range,
                    reduction="none",
                )
                .detach()
                .cpu()
                .item()
                for img in imgs
            ]

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": scores},
        )


__all__ = [
    "Tenengrad",
    "TenengradRelative",
    "EME",
    "ExposureBrightness",
    "Brightness",
    "Colorfulness",
    "CLIPIQA",
    "BRISQUE",
]



from typing import Optional

import numpy as np
import spacy
import torch
from loguru import logger
from sklearn.decomposition import PCA
from sklearn.manifold import Isomap, LocallyLinearEmbedding
from transformers import BertModel, BertTokenizer
from umap.umap_ import UMAP

from ..embeddings import GenericEmbedder
from ..embeddings.utils import tabular_to_text


class UMAPEmbedder(GenericEmbedder):
    """UMAP (Uniform Manifold Approximation and Projection) embedder using
    default and customizable parameters for dimensionality reduction.

    Example
    -------
    >>> import numpy as np
    >>> embedder = UMAPEmbedder(n_neighbors=10, n_components=2)
    >>> X = np.random.rand(100, 10)
    >>> X_embed = embedder.fit_transform(X)
    >>> print(X_embed)
    """

    D_PARAMS = {
        "n_neighbors": 15,
        "n_components": 3,
        "metric": "euclidean",
        "output_metric": "euclidean",
        "n_epochs": None,
        "learning_rate": 1.0,
        "init": "spectral",
        "min_dist": 0.1,
        "spread": 1.0,
        "low_memory": True,
        "n_jobs": 1,
        "random_state": 42,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # embedder params
        emb_params = {**self.D_PARAMS, **kwargs}

        # initialize
        try:
            # try instance with provided params
            self.emb_obj = UMAP(**emb_params)
        except Exception as e:
            logger.debug(f"Error:\n{e}")

            # default instance (in case of any error)
            self.emb_obj = UMAP(**self.D_PARAMS)

        # has inverse transform
        self.has_inv = True

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
        force_all_finite: bool = "allow-nan",
        **kwargs,
    ) -> None:
        self.emb_obj.fit(X, y, force_all_finite)
        return self

    def transform(
        self,
        X: np.ndarray,
        force_all_finite: bool = "allow-nan",
        **kwargs,
    ) -> np.ndarray:
        return self.emb_obj.transform(X, force_all_finite)

    def fit_transform(
        self,
        X: np.ndarray,
        y: np.ndarray = None,
        force_all_finite: bool = "allow-nan",
        **kwargs,
    ) -> np.ndarray:
        return self.emb_obj.fit_transform(X, y, force_all_finite)

    def inverse_transform(self, X, **kwargs):
        return self.emb_obj.inverse_transform(X)


class PCAEmbedder(GenericEmbedder):
    """PCA (Principal Component Analysis) embedder for dimensionality
    reduction.

    Example
    -------
    >>> import numpy as np
    >>> embedder = PCAEmbedder(n_components=5)
    >>> X = np.random.rand(100, 10)
    >>> X_embed = embedder.fit_transform(X)
    >>> print(X_embed)
    """

    D_PARAMS = {
        "n_components": None,
        "copy": True,
        "whiten": False,
        "svd_solver": "auto",
        "tol": 0.0,
        "iterated_power": "auto",
        "n_oversamples": 10,
        "power_iteration_normalizer": "auto",
        "random_state": 42,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # embedder params
        emb_params = {**self.D_PARAMS, **kwargs}

        # initialize
        try:
            self.emb_obj = PCA(**emb_params)
        except Exception as e:
            print(f"Error: {e}")
            self.emb_obj = PCA(**self.D_PARAMS)

        # has inverse transform
        self.has_inv = True

    def fit(self, X, y=None, **kwargs):
        self.emb_obj.fit(X, y)
        return self

    def transform(self, X, **kwargs):
        return self.emb_obj.transform(X)

    def fit_transform(self, X, **kwargs):
        return self.emb_obj.fit_transform(X)

    def inverse_transform(self, X, **kwargs):
        return self.emb_obj.inverse_transform(X)


class LLEEmbedder(GenericEmbedder):
    """Locally Linear Embedding (LLE) embedder for non-linear dimensionality
    reduction.

    Example
    -------
    >>> import numpy as np
    >>> embedder = LLEEmbedder(n_neighbors=10, n_components=2)
    >>> X = np.random.rand(100, 10)
    >>> X_embed = embedder.fit_transform(X)
    >>> print(X_embed)
    """

    D_PARAMS = {
        "n_neighbors": 5,
        "n_components": 2,
        "reg": 0.001,
        "eigen_solver": "dense",
        "random_state": 42,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # embedder params
        emb_params = {**self.D_PARAMS, **kwargs}

        # initialize
        try:
            self.emb_obj = LocallyLinearEmbedding(**emb_params)
        except Exception as e:
            logger.debug(f"Error:\n{e}")
            self.emb_obj = LocallyLinearEmbedding(**self.D_PARAMS)

        # has inverse transform
        self.has_inv = False

    def fit(self, X, y=None, **kwargs):
        self.emb_obj.fit(X, y)
        return self

    def transform(self, X, **kwargs):
        return self.emb_obj.transform(X)

    def fit_transform(self, X, y=None, **kwargs):
        return self.emb_obj.fit_transform(X, y)


class IsomapEmbedder(GenericEmbedder):
    """Isomap embedder for manifold learning and dimensionality reduction.

    Example
    -------
    >>> import numpy as np
    >>> embedder = IsomapEmbedder(n_neighbors=5, n_components=3)
    >>> X = np.random.rand(100, 10)
    >>> X_embed = embedder.fit_transform(X)
    >>> print(X_embed)
    """

    D_PARAMS = {
        "n_neighbors": 6,
        "radius": None,
        "n_components": 3,
        "eigen_solver": "auto",
        "tol": 0,
        "max_iter": None,
        "path_method": "auto",
        "neighbors_algorithm": "auto",
        "n_jobs": -1,
        "metric": "minkowski",
        "p": 2,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # embedder params
        emb_params = {**self.D_PARAMS, **kwargs}

        # initialize
        try:
            self.emb_obj = Isomap(**emb_params)
        except Exception as e:
            print(f"Error: {e}")
            self.emb_obj = Isomap(**self.D_PARAMS)

        # has inverse transform
        self.has_inv = False

    def fit(self, X, **kwargs):
        self.emb_obj.fit(X)
        return self

    def transform(self, X, **kwargs):
        return self.emb_obj.transform(X)

    def fit_transform(self, X, **kwargs):
        return self.emb_obj.fit_transform(X)


class SpacyWordEmbedder(GenericEmbedder):
    """Word embedding model using spaCy for text data transformation into
    vectors.

    Parameters
    ----------
    model_name : str
        The name of the spaCy model to use for word embeddings.

    Example
    -------
    >>> import numpy as np
    >>> embedder = SpacyWordEmbedder(model_name='en_core_web_md')
    >>> X = np.random.rand(100, 10)
    >>> X_embed = embedder.transform(X)
    >>> print(X_embed)
    """

    def __init__(self, model_name: Optional[str] = "en_core_web_md"):
        super().__init__()
        try:
            # try to load model (check if exists)
            self.emb_obj = spacy.load(model_name)
        except (SystemExit, SystemError, OSError) as e:
            logger.debug(f"Error:\n{e}")
            # download it
            spacy.cli.download(model_name)

            # now load it again
            self.emb_obj = spacy.load(model_name)

        # has inverse transform
        self.has_inv = False

    def table_to_text(self, X, **kwargs):
        # records to prompt
        prompts = tabular_to_text(X, **kwargs)
        return prompts

    def fit(self, X, y=None, **kwargs):
        # word embeddings are pre-trained,
        # so no fitting needed
        return self

    def transform(self, X, **kwargs):
        # prompt-encoded data
        logger.debug(X)
        X_ = self.table_to_text(X, **kwargs)

        # get embeddings
        embeddings = [self.emb_obj(sentence).vector for sentence in X_]

        return np.array(embeddings)

    def fit_transform(self, X, **kwargs):
        return self.transform(X, **kwargs)


class BertSentenceEmbedder(GenericEmbedder):
    """BERT-based sentence embedder for transforming text data into dense
    vectors.

    Parameters
    ----------
    model_name : str
        The name of the BERT model to use for sentence embeddings.

    Example
    -------
    >>> import numpy as np
    >>> embedder = BertSentenceEmbedder()
    >>> X = np.random.rand(100, 10)
    >>> X_embed = embedder.transform(X)
    >>> print(X_embed)
    """

    def __init__(self, model_name: Optional[str] = "bert-base-uncased"):
        super().__init__()
        try:
            self.tokenizer = BertTokenizer.from_pretrained(model_name)
            self.emb_obj = BertModel.from_pretrained(model_name)
        except Exception as e:
            logger.debug(f"Error:\n{e}")
            self.tokenizer = BertTokenizer.from_pretrained("bert-base-uncased")
            self.emb_obj = BertModel.from_pretrained("bert-base-uncased")

        # has inverse transform
        self.has_inv = False

    def table_to_text(self, X, **kwargs):
        # encode tabular rows into text prompts
        table_prompts = tabular_to_text(data=X, **kwargs)

        return table_prompts

    def fit(self, X, y=None, **kwargs):
        # sentence embeddings are pre-trained, so no fitting needed
        return self

    def transform(self, X, **kwargs):
        # prompt-encoded data
        X_ = self.table_to_text(X, **kwargs)

        # tokenize input
        tokenized = self.tokenizer(
            X_,
            padding=True,
            truncation=True,
            return_tensors="pt",
        )

        # get embeddings
        with torch.no_grad():
            outputs = self.emb_obj(**tokenized)
        return outputs.last_hidden_state.mean(dim=1).numpy()

    def fit_transform(self, X, **kwargs):
        return self.transform(X, **kwargs)


class GlobalEmbedder:
    """Global embedder that selects and applies the appropriate embedding
    method.

    Parameters
    ----------
    n_type : Optional[str]
        The type of embedding method to use.
        Must be one of: "umap", "pca", "lle", "isomap", "spacy", "bert", "identity".

    Example
    -------
    >>> import numpy as np
    >>> embedder = GlobalEmbedder(n_type="pca")
    >>> X = np.random.rand(100, 10)
    >>> X_embed = embedder.fit_transform(X)
    >>> print(X_embed)
    """

    E_MAP = {
        "umap": UMAPEmbedder,
        "pca": PCAEmbedder,
        "lle": LLEEmbedder,
        "isomap": IsomapEmbedder,
        "spacy": SpacyWordEmbedder,
        "bert": BertSentenceEmbedder,
        "identity": GenericEmbedder,
    }

    def __init__(self, n_type: Optional[str] = "umap", **kwargs):
        # get embedding object
        if isinstance(n_type, str) and n_type in self.E_MAP:
            # tag
            norm_tag = n_type.lower()

            # embed class
            cls_ = self.E_MAP.get(norm_tag, GenericEmbedder)

            # embed instance
            self.emb_obj = cls_(**kwargs)
        else:
            # default embed instance
            self.emb_obj = GenericEmbedder()

        # has inverse transform
        self.has_inv = self.emb_obj.has_inv

    def fit(self, X, **kwargs):
        self.emb_obj.fit(X, **kwargs)
        return self

    def transform(self, X, **kwargs):
        return self.emb_obj.transform(X, **kwargs)

    def fit_transform(self, X, **kwargs):
        return self.emb_obj.fit_transform(X, **kwargs)

    def inverse_transform(self, X, **kwargs):
        return self.emb_obj.inverse_transform(X, **kwargs)



from typing import List, Tuple

import cv2
import numpy as np


def downsample_largest(im_1, im_2):
    size_1 = im_1.shape[0] * im_1.shape[1]
    size_2 = im_2.shape[0] * im_2.shape[1]

    if size_1 > size_2:
        return cv2.resize(im_1, (im_2.shape[1], im_2.shape[0]), interpolation=cv2.INTER_LANCZOS4), im_2
    return im_1, cv2.resize(im_2, (im_1.shape[1], im_1.shape[0]), interpolation=cv2.INTER_LANCZOS4)


def batch_downsample_to_largest(imgs: List[np.ndarray], imgs2: List[np.ndarray]) -> Tuple[List[np.ndarray]]:
    """Downsamples the largest image in the image pair.

    Args:
        imgs (List[np.ndarray]): left images
        imgs2 (List[np.ndarray]): right images

    Returns:
        Tuple[np.ndarray]: tuple with the image lists
    """
    resized = [downsample_largest(img, img2) for img, img2 in zip(imgs, imgs2)]
    return [img for img, _ in resized], [img2 for _, img2 in resized]


def image_resize(
    image: np.ndarray,
    height: int = None,
    width: int = None,
    inter: int = cv2.INTER_LANCZOS4,
) -> np.ndarray:
    """Resize an image to a given width or height.

    Aspect ratio is kept if only one dimmension is given.
    """
    # initialize the dimensions of the image to be resized and
    # grab the image size
    dim = None
    (h, w) = image.shape[:2]

    # if both the width and height are None, then return the
    # original image
    if width is None and height is None:
        return image

    if width is not None and height is not None:
        return cv2.resize(image, (width, height), interpolation=inter)

    # check to see if the width is None
    if width is None:
        # calculate the ratio of the height and construct the
        # dimensions
        r = height / float(h)
        dim = (int(w * r), height)

    # otherwise, the height is None
    else:
        # calculate the ratio of the width and construct the
        # dimensions
        r = width / float(w)
        dim = (width, int(h * r))

    # resize the image
    resized = cv2.resize(image, dim, interpolation=inter)

    # return the resized image
    return resized
